"""
Module containing dashboard elements.
"""
from datetime import datetime, timedelta
from threading import Lock, Thread
from typing import Any, Dict, List, Type

from .base_control import DialogFieldResourceListSelector, OpenableListControl
from .common import Common, SessionAwareDialog, default_args, default_kwargs
from .hotkey import HotkeyDisplay
from .termui.alignment import Dimension, TopLeftDimensionAnchor
from .termui.control import HotkeyControl
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes


class Dashboard(HotkeyControl):
    """
    Dashboard controller class. The dashboard contains a number of dashboard blocks organized in a table.
    """

    title = "Dashboard"
    prefix = "dashboard"

    block_registry: Dict[str, Type] = {}

    def __init__(self, *args, layout, **kwargs):
        super().__init__(*args, **kwargs)
        if len(layout) == 0:
            return
        self.refresh_loop = []
        self.load_layout(layout)
        self.add_hotkey(ControlCodes.R, self.force_refresh, "Refresh")
        self.add_hotkey("d", self.configure_default_layout, "Configure default")
        self.add_hotkey("c", self.configure_context_layout, "Configure context")
        self.hotkey_display = HotkeyDisplay.opener(caller=self)

    def reload_layout(self):
        """
        Callback for the configuration dialog to reload the dashboard.
        """
        self.load_layout(Common.Session.get_context_dashboard_layout())

    def configure_default_layout(self, *args):
        """
        Hotkey callback for configuring the default dashboard layout.
        """
        DashboardConfigurationDialog.opener(caller=self, is_default=True)

    def configure_context_layout(self, *args):
        """
        Hotkey callback for configuring the context dashboard layout.
        """
        DashboardConfigurationDialog.opener(caller=self, is_default=False)

    def load_layout(self, layout):
        """
        Reloads the dashboard with the layout configuration provided.

        Parameters
        ----------
        layout : list(list(str or class))
            The dashboard layout to load.
        """
        self.refresh_loop = []
        self.clear_blocks()
        h_perc = int(100.0 / len(layout))
        y = 0
        for row in layout:
            if len(row) == 0:
                continue
            w_perc = int(100.0 / len(row))
            x = 0
            for elem in row:
                cls = elem
                if isinstance(elem, str):
                    cls = Dashboard.block_registry[elem]
                self.refresh_loop.append(
                    cls(
                        self,
                        TopLeftDimensionAnchor(f"{w_perc * x}%", f"{h_perc * y}%"),
                        Dimension(f"{w_perc}%", f"{h_perc}%"),
                    )
                )
                x += 1
            y += 1

    def force_refresh(self, *args):
        """
        Forcibly refreshes blocks. Used for user input.
        """
        for elem in self.refresh_loop:
            elem.auto_refresh_data(True)

    def auto_refresh(self):
        """
        auto_refresh is called on top level blocks that have it automatically by the main loop.

        Checks periodically if subblocks need refreshing.
        """
        for elem in self.refresh_loop:
            elem.auto_refresh_data()

    @classmethod
    def opener(cls, layout=None, **kwargs):
        """
        Session-aware initializer for this class. Creates an ResourceLister object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.OpenableListControl, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        if layout is None:
            layout = Common.Session.get_context_dashboard_layout()
        instance = cls(
            *default_args(), **default_kwargs(cls.prefix), layout=layout, **kwargs
        )
        instance.border = Common.default_border(cls.prefix, cls.title, None)
        return [instance, instance.hotkey_display]

    @classmethod
    def register(cls):
        """
        Registers everything related to dashboards.
        """
        DashboardBlock.register()
        Common.Session.commander_options["dashboard"] = cls.opener
        Common.Session.commander_options["db"] = cls.opener


class DashboardConfigurationDialog(SessionAwareDialog):
    """
    Modal dialog for configuring the dashboard.
    """

    line_size = 20

    def __init__(self, *args, is_default=False, **kwargs):
        kwargs["border"] = Common.default_border("modal_dialog", "Layout configuration")
        super().__init__(*args, **kwargs)
        self.is_default = is_default

        layout = (
            Common.Session.get_default_dashboard_layout()
            if is_default
            else Common.Session.get_context_dashboard_layout()
        )
        if is_default:
            self.set_title_label("Configure layout for default dashboard")
        else:
            self.set_title_label(
                f"Configure dashboard layout for context {Common.Session.context}"
            )

        self.selectors = []

        # Currently, we assume all dashboards are 2x2. Later, this may change.
        for row in range(2):
            self.selectors.append([])
            for col in range(2):
                field = DialogFieldResourceListSelector(
                    DashboardBlockLister,
                    f"Dashboard row {row} block {col}",
                    default=layout[row][col],
                    primary_key="name",
                )
                self.selectors[row].append(field)
                self.add_field(field)

    def accept_and_close(self):
        config = Common.Session.config
        for row in self.selectors:
            for col in row:
                if col.value == "":
                    col.value = "Blank"
        if self.is_default:
            config["default_dashboard_layout"] = [
                [selector.value for selector in row] for row in self.selectors
            ]
        else:
            config["dashboard_layouts"][Common.Session.context] = [
                [selector.value for selector in row] for row in self.selectors
            ]
        config.write_config()
        Common.Session.set_message(
            "Successfully updated dashboard layout.", Common.color("message_success")
        )
        self.caller.reload_layout()
        super().accept_and_close()


class DashboardBlockLister(OpenableListControl):
    """
    Lister control for selecting dashboard blocks to add.
    """

    prefix = "dashboard_block"
    title = "Available Dashboard Blocks"

    def __init__(self, *args, selector_cb=None, **kwargs):
        super().__init__(
            *args,
            color=Common.color("dashboard_block_list_generic", "generic"),
            selection_color=Common.color("dashboard_block_list_selection", "selection"),
            title_color=Common.color("dashboard_block_list_heading", "column_title"),
            selector_cb=selector_cb,
            **kwargs,
        )

        self.add_column("classname", 12)
        self.add_column("description", 60)
        self.reload()

    def reload(self):
        """
        Reloads the list of available dashboard blocks.
        """
        self.entries = []
        for block, block_class in Dashboard.block_registry.items():
            if (
                block in ("DashboardBlock", "KeyValueDashboardBlock")
                or block_class.description == ""
            ):
                continue
            self.add_entry(
                ListEntry(
                    block,
                    **{
                        "classname": block,
                        "description": block_class.description,
                    },
                )
            )


class DashboardBlock(HotkeyControl):
    """
    Parent class for individual dashboard blocks. The intent of dashboard blocks is to display condensed information about a particular AWS service.

    Attributes
    ----------
    refresh_frequency : int
        How often (in seconds) to refresh the data displayed on the dashboard block automatically.
    """

    description = ""
    refresh_frequency = 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._refresh_frequency = timedelta(seconds=self.refresh_frequency)
        self.mutex = Lock()
        self.refreshing = False
        self.last_refresh = datetime.now()
        self.async_refresh()

    def async_refresh(self):
        """
        Async flow for refresh_data.
        """
        if self.refreshing:
            return
        self.refreshing = True
        thread = Thread(
            target=self._async_refresh_wrapper,
            daemon=True,
        )
        thread.start()

    def _async_refresh_wrapper(self):
        self.refresh_data()
        self.refreshing = False

    def auto_refresh_data(self, force=False):
        """
        Called by Dashboard to facilitate auto-refreshing.

        Parameters
        ----------
        force : bool, default=False
            Bypasses time checks if True.
        """
        if force or datetime.now() - self.last_refresh > self._refresh_frequency:
            self.async_refresh()
            self.last_refresh = datetime.now()

    def refresh_data(self):
        """
        Called automatically to refresh the data on the dashboard block. Subclasses are expected to override this.
        """

    @classmethod
    def register(cls):
        """
        Classmethod to register this class and all of its subclasses with the commander control.

        Also stores the class in the session control registry for other uses.

        This is why we can reach things via the command palette.
        """
        Dashboard.block_registry[cls.__name__] = cls
        for subcls in cls.__subclasses__():
            subcls.register()


class KeyValueDashboardBlock(DashboardBlock):
    """
    Key-value display template for dashboard blocks. Blocks that do nothing fancy, just display info, should use this.

    Attributes
    ----------
    info : dict
        A mapping of fields to field values. Refreshing the data should update this.
    status : int
        The current status of the block, comparable with the STATUS constants. Handling the status field is the responsibility of the
        subclass implementing refresh_data.
    labels : dict
        A mapping of fields to field labels for display purposes. Only fields present in labels, order and info will be rendered.
    order : list
        A list of fields. The order in which fields should be rendered.
    thresholds : dict
        A mapping of fields to threshold definitions. A threshold definition is a list alternating color names and numeric values. The first and the last element
        must always be a color. The color represents the color to print the value with if it's less than the next numeric value in the list, but greater than the
        previous numeric value of the list. There is an assumed negative infinity at the beginning, and positive infinity at the end of the list.
    suffixes : dict
        A mapping of field names to suffixes to render after the field value.
    inverted : list
        A list of field names where the label should be printed after the field value.
    additional_lines : list(tuple)
        A list of additional lines to display at the end of the block. Each entry in the list is a tuple in the format of (str, str, bool), where the first entry
        is the line to display, the second is the name of the color for the line, and the third is whether the line should be bold.
    """

    STATUS_LOADING = 0
    """
    Status value which represents that the block data is still loading.
    """

    STATUS_ERROR = 1
    """
    Status value which represents that the block data failed to load in the last refresh cycle.
    """

    STATUS_READY = 2
    """
    Status value which represents that the block data succeeded to load in the last refresh cycle.
    """

    labels: Dict[str, str] = {}
    order: List[str] = []
    thresholds: Dict[str, List[Any]] = {}
    suffixes: Dict[str, str] = {}
    inverted: List[str] = []

    def __init__(self, *args, **kwargs):
        self.info = {}
        self.additional_lines = []
        self.status = self.STATUS_LOADING
        super().__init__(*args, **kwargs)

    def paint(self):
        super().paint()
        with self.mutex:
            bounds = self.corners
            Common.Session.ui.print(
                self.description,
                xy=(bounds[0][0] + 1, bounds[1][0] + 1),
                bounds=bounds,
                color=Common.color("dashboard_block_label"),
                bold=True,
            )
            if self.status == self.STATUS_LOADING:
                Common.Session.ui.print_centered(
                    "Loading...",
                    bounds,
                    color=Common.color("dashboard_block_loading"),
                    bold=True,
                )
            elif self.status == self.STATUS_ERROR:
                Common.Session.ui.print_centered(
                    "Error loading data!",
                    bounds,
                    color=Common.color("dashboard_block_error"),
                    bold=True,
                )
            else:
                y = bounds[1][0] + 3
                for field in self.order:
                    if field not in self.info or field not in self.labels:
                        continue
                    x = bounds[0][0] + 1
                    if field not in self.inverted:
                        output = f"{self.labels[field]}: "
                        Common.Session.ui.print(
                            output,
                            xy=(x, y),
                            bounds=bounds,
                            color=Common.color("dashboard_block_label"),
                            bold=True,
                        )
                        x += len(output)
                    if field in self.thresholds:
                        is_color = True
                        for elem in self.thresholds[field]:
                            if is_color:
                                color = Common.color(elem)
                            else:
                                if self.info[field] < elem:
                                    break
                            is_color = not is_color
                    else:
                        color = Common.color("dashboard_block_information")
                    output = str(self.info[field])
                    if field in self.suffixes:
                        output = f"{output}{self.suffixes[field]}"
                    Common.Session.ui.print(
                        output,
                        xy=(x, y),
                        bounds=bounds,
                        color=color,
                        bold=False,
                    )
                    x += len(output)
                    if field in self.inverted:
                        output = f" {self.labels[field]}"
                        Common.Session.ui.print(
                            output,
                            xy=(x, y),
                            color=Common.color("dashboard_block_label"),
                            bold=True,
                            bounds=bounds,
                        )
                    y += 1
                for line in self.additional_lines:
                    x = bounds[0][0] + 1
                    output = line[0]
                    color = Common.color(line[1])
                    bold = line[2]
                    Common.Session.ui.print(
                        output,
                        xy=(x, y),
                        color=color,
                        bold=bold,
                        bounds=bounds,
                    )
                    y += 1

                if self.refreshing:
                    Common.Session.ui.print_centered(
                        "Refreshing...",
                        bounds,
                        color=Common.color("dashboard_block_loading"),
                        bold=True,
                    )


class Blank(DashboardBlock):
    """
    Alias for DashboardBlock. Does nothing, just occupies space.
    """

    description = "Blank space"

    def paint(self):
        super().paint()
        Common.Session.ui.print_centered(
            "Empty dashboard block",
            self.corners,
            color=Common.color("dashboard_block_label"),
            bold=True,
        )



"""
Module that contains the special control bars for the awsc UI.
"""
from .termui.common import Commons
from .termui.control import Control


class Filterer(Control):
    """
    The Filterer is the command bar for filtering the currently active control. Different top level controls react differently to filtering.
    Lister controls will hide elements not matching the filter. Text browser controls will highlight matches in their text and jump between
    matches.

    Attributes
    ----------
    session : awsc.session.Session
        The session manager object.
    color : awsc.termui.color.Color
        The color of the control's text and borders.
    symbol_color : awsc.termui.color.Color
        The color of the symbol for the filter bar.
    inactive_color : awsc.termui.color.Color
        The color of the bar when it is inactive.
    paused : bool
        Whether the bar is inactive. The inactive bar still applies its filter, but pressing keys does not type into it.
    accepted_input : str
        The list of accepted input keys. Defaults to accepted textfield inputs.
    text : str
        The current filter.
    """

    def __init__(self, *args, session, color, symbol_color, inactive_color, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session
        self.color = color
        self.symbol_color = symbol_color
        self.inactive_color = inactive_color
        self.paused = False
        self.session.filterer = self
        self.accepted_input = Commons.TextfieldInputs
        self.text = ""

    def input(self, key):
        if self.paused and not (key.is_sequence and key.name == "KEY_ESCAPE"):
            return False
        if key.is_sequence:
            if key.name == "KEY_ENTER":
                self.pause()
            elif key.name == "KEY_ESCAPE":
                self.session.resource_main.filter = None
                self.close()
            elif key.name == "KEY_BACKSPACE":
                if len(self.text) > 0:
                    self.text = self.text[:-1]
                self.session.resource_main.filter = self.text
            elif key.name == "KEY_DELETE":
                self.text = ""
                self.session.resource_main.filter = self.text
            Commons.UIInstance.dirty = True
        elif str(key) in self.accepted_input:
            self.text += key
            self.session.resource_main.filter = self.text
            Commons.UIInstance.dirty = True
        return True

    def pause(self):
        """
        Sets pause flag.
        """
        self.paused = True

    def resume(self):
        """
        Unsets pause flag.

        Returns
        -------
        awsc.commander.Filterer
            Self.
        """
        self.paused = False
        return self

    def close(self):
        """
        Closes the filter bar and removes it from view.
        """
        self.parent.remove_block(self)
        self.paused = True
        self.session.filterer = None

    def paint(self):
        super().paint()
        (x0, x1), (y0, _) = self.inner
        Commons.UIInstance.print("/", xy=(x0 + 1, y0), color=self.symbol_color)
        space = (x1 - 1) - (x0 + 2) + 1
        text = self.text if len(self.text) <= space else self.text[:space]
        Commons.UIInstance.print(
            text,
            xy=(x0 + 2, y0),
            color=self.color if not self.paused else self.inactive_color,
        )


class Commander(Control):
    """
    The Commander is the command bar for switching between controls. It allows navigation between AWS resources by typing commands defined
    by the command_palette attribute of ResourceListers into the command bar.

    Attributes
    ----------
    session : awsc.session.Session
        The session manager object.
    color : awsc.termui.color.Color
        The color of the control's text and borders.
    symbol_color : awsc.termui.color.Color
        The color of the symbol for the filter bar.
    autocomplete_color : awsc.termui.color.Color
        The color of the autocompletion text in the bar.
    ok_color : awsc.termui.color.Color
        The color of the text in the bar when it fully matches a valid command.
    autocomplete_color : awsc.termui.color.Color
        The color of the text in the bar when it does not match a valid command and does not autocomplete to a valid command.
    paused : bool
        Whether the bar is inactive. The inactive bar still applies its filter, but pressing keys does not type into it.
    accepted_input : str
        The list of accepted input keys. Defaults to accepted textfield inputs.
    text : str
        The current filter.
    options : dict(str, callable() -> list(awsc.termui.control.Control))
        A mapping of valid options to their respective callbacks.
    """

    def __init__(
        self,
        *args,
        session,
        color,
        symbol_color,
        autocomplete_color,
        ok_color,
        error_color,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.session = session
        self.color = color
        self.symbol_color = symbol_color
        self.autocomplete_color = autocomplete_color
        self.ok_color = ok_color
        self.error_color = error_color
        self.session.commander = self
        self.accepted_input = Commons.TextfieldInputsAlphaNum
        self.text = ""
        self.options = session.commander_options

    def autocomplete(self):
        """
        Autocompletion callback. Returns the first autocompletion match for the text that has been typed. The first match is always the
        shortest possible match.

        Returns
        -------
        str
            The autocompleted command. None if nothing has been typed or no matches are found.
        """
        if len(self.text) == 0:
            return None
        opt_keys = self.options.keys()
        candidates = [cand for cand in opt_keys if cand.startswith(self.text.lower())]
        if len(candidates) > 0:
            candidates.sort(key=len)
            return candidates[0]
        return None

    def accept_autocomplete(self):
        """
        Called when the user accepts the autocompletion, by pressing TAB. Fills the command bar to the autocompletion.
        """
        acp = self.autocomplete()
        if acp is not None:
            self.text = acp

    def input(self, key):
        if key.is_sequence:
            if key.name == "KEY_ENTER":
                self.accept_and_close()
            elif key.name == "KEY_ESCAPE":
                self.close()
            elif key.name == "KEY_BACKSPACE":
                if len(self.text) > 0:
                    self.text = self.text[:-1]
            elif key.name == "KEY_DELETE":
                self.text = ""
            elif key.name == "KEY_TAB":
                self.accept_autocomplete()
            Commons.UIInstance.dirty = True
        elif str(key) in self.accepted_input:
            self.text += key
            Commons.UIInstance.dirty = True
        return True

    def accept_and_close(self):
        """
        Hotkey callback for pressing ENTER. If the text is a valid command in the bar, opens the associated control. If the command is not
        valid but autocompletes to a valid command, that command is used instead. If neither is valid, does nothing.
        """
        text = self.text.lower()
        option = None
        if text in self.options:
            option = self.options[text]
        else:
            acp = self.autocomplete()
            if acp is not None:
                option = self.options[acp]
        if option is not None:
            self.session.replace_frame(option())
            self.close()

    def close(self):
        """
        Closes the commander and removes it from view.
        """
        self.parent.remove_block(self)
        self.session.commander = None

    def paint(self):
        super().paint()
        (x0, x1), (y0, _) = self.inner
        Commons.UIInstance.print(":", xy=(x0 + 1, y0), color=self.symbol_color)
        space = (x1 - 1) - (x0 + 2) + 1
        text = self.text if len(self.text) <= space else self.text[:space]
        acp = self.autocomplete()
        color = self.color
        if self.text.lower() in self.options:
            color = self.ok_color
        elif acp is None:
            color = self.error_color
        Commons.UIInstance.print(text, xy=(x0 + 2, y0), color=color)
        if acp is not None:
            acp = acp[len(self.text) :]
            if len(acp) > 0:
                Commons.UIInstance.print(
                    acp, xy=(x0 + 2 + len(self.text), y0), color=self.autocomplete_color
                )



"""
Module for working with the AWS API.
"""

import boto3
from botocore import config as botoconf
from botocore import exceptions as botoerror

from .common import Common


class AWS:
    """
    Class which bridges the boto3 class model with the awsc configuration model.
    """

    def __init__(self):
        """
        Initializes an AWS object.
        """
        Common.Session.context_update_hooks.append(self.idcaller)
        self.idcaller()

    def conf(self):
        """
        Generates a boto3 configuration object from the current awsc session.

        Returns
        -------
        botocore.config.Config
            The configuration object with the region and signature version set.
        """
        return botoconf.Config(
            region_name=Common.Session.region,
            signature_version="v4",
        )

    def s3conf(self):
        """
        Generates a boto3 configuration object for s3 from the current awsc session.

        Returns
        -------
        botocore.config.Config
            The configuration object with the region and signature version set.
        """
        return botoconf.Config(
            region_name=Common.Session.region,
            signature_version="s3v4",
        )

    def env_session(self):
        """
        Fetches the boto3 session for the AWS credentials configured in the process environment.

        Returns
        -------
        boto3.session.Session
            The session object corresponding to the current process environment.
        """
        return boto3.Session()

    def __call__(self, service, keys=None):
        """
        Shorthand for boto3.client().

        Parameters
        ----------
        service : str
            The service parameter for boto3.client()
        keys : dict
            A dict-like object with the "access" and "secret" keys set. If None, the currently
            active keypair in the configuration is used.

        Returns
        -------
        object
            A boto3 client instance for the service.
        """
        if service == "s3":
            config = self.s3conf()
        else:
            config = self.conf()
        if keys is None:
            if Common.Session.context not in Common.Configuration.keystore:
                return boto3.client(service)  # Let magic sort it out
            keys = Common.Configuration.keystore[Common.Session.context]
        access = keys["access"]
        secret = keys["secret"]
        endpoint = None
        session = None
        if "session" in keys:
            session = keys["session"]
        if "endpoint_url" in Common.Configuration["contexts"][Common.Session.context]:
            endpoint = Common.Configuration["contexts"][Common.Session.context][
                "endpoint_url"
            ]
        # print(f'Connecting to aws with keypair {access} / {secret} for context {Common.Session.context} on ')
        client = boto3.client(
            service,
            aws_access_key_id=access,
            aws_secret_access_key=secret,
            aws_session_token=session,
            config=config,
            endpoint_url=endpoint,
        )
        # return AWSSubprocessWrapper(client)
        return client

    # def assume_role(
    #     self,
    #     target_role,
    #     security_token="",
    #     mfa_device="",
    #     token="",
    #     session_name="awsc-session",
    #     duration=3600,
    #     keys=None,
    # ):
    #     if keys is None:
    #         if Common.Session.context not in Common.Configuration.keystore:
    #             access = None
    #             secret = None
    #         keys = Common.Configuration.keystore.force_resolve(Common.Session.context)
    #     access = keys["access"]
    #     secret = keys["secret"]
    #     endpoint = None
    #     if "endpoint_url" in Common.Configuration["contexts"][Common.Session.context]:
    #         endpoint = Common.Configuration["contexts"][Common.Session.context][
    #             "endpoint_url"
    #         ]
    #     if security_token == "" and mfa_device != "":
    #
    #         sts = boto3.client(
    #             "sts",
    #             aws_access_key_id=access,
    #             aws_secret_access_key=secret,
    #             config=self.conf(),
    #             endpoint=endpoint,
    #         )
    #         sts.get_session_t

    def whoami(self, keys=None):
        """
        Shorthand for the GetCallerIdentity STS API call.

        Parameters
        ----------
        keys : dict
            A dict-like object with the "access" and "secret" keys set. If None, the currently
            active keypair in the configuration is used.

        Returns
        -------
        object
            The API response for GetCallerIdentity.
        """
        return self("sts", keys).get_caller_identity()

    def list_regions(self):
        """
        Shorthand for enumerating the result of the DescribeRegions EC2 API call.

        Returns
        -------
        list
            A list of valid region names for EC2.
        """
        return [
            region["RegionName"]
            for region in self("ec2").describe_regions(AllRegions=True)["Regions"]
        ]

    def idcaller(self):
        """
        Sets the info display for the account and user ID of the currently selected keypair.

        AWSC must be fully initialized before calling.
        """
        try:
            whoami = self.whoami()
            try:
                del Common.Session.info_display.special_colors["Account"]
            except KeyError:
                pass
            try:
                del Common.Session.info_display.special_colors["UserId"]
            except KeyError:
                pass
            Common.Session.info_display["UserId"] = whoami["UserId"]
            Common.Session.info_display["Account"] = whoami["Account"]
        except (botoerror.ClientError, KeyError) as error:
            Common.Session.info_display.special_colors["UserId"] = Common.color("error")
            Common.Session.info_display.special_colors["Account"] = Common.color(
                "error"
            )
            Common.Session.info_display["UserId"] = "ERROR"
            Common.Session.info_display["Account"] = "ERROR"
            if isinstance(error, botoerror.ClientError):
                Common.clienterror(
                    error,
                    "Identify caller",
                    "Core",
                    subcategory="STS",
                )
            else:
                Common.error(
                    str(error),
                    "Identify caller",
                    "Core",
                    subcategory="STS",
                )



"""
This module defines classes for aligning blocks on the terminal.
"""
from .common import Commons


class Dimension:
    """
    Represents a width and height value.

    Upon each axis, the value can be specified in a few ways:
    * Directly, as a number of characters. (eg. 20)
    * As a percentage of space available in the current block. (eg. 100%)
    * As a difference of any amount of the above values (eg. 100%-5, 100%-20%, 5-2, 100%-2-1)
    * As a sum of any amount of the above values (eg. 50%+1)
    * As the highest character count of any number of above operations (eg. 50%|60%-5|20)

    Attributes
    ----------
    width : str
        The width specification.
    height : str
        The height specification.
    """

    def __init__(self, width, height):
        """
        Initializes a Dimension object.

        Parameters
        ----------
        width : str
            The width specification.
        height : str
            The height specification.
        """
        self.width = width
        self.height = height

    def __getitem__(self, key):
        """
        Shorthand for retrieving the calculated width or height.

        Parameters
        ----------
        key : int|str
            One of 0, 1, "x", "y", "w" or "h".

        Returns
        -------
        int
            The evaluated width for keys 0, "x" or "w", otherwise, the evaluated height.
        """
        if key not in (0, 1, "x", "y", "w", "h"):
            raise KeyError("Bad dimension key " + key)
        val = self()
        return val[0] if key in (0, "x", "w") else val[1]

    @staticmethod
    def _v(value, space):
        """
        Evaluator function. Parses an axis specification and returns the character cell count.

        Parameters
        ----------
        value : str
            An axis specification.
        space : int
            The total space available on the axis.

        Returns
        -------
        int
            The resulting dimension.
        """
        if isinstance(value, str):
            test = value.split("|")
            if len(test) > 1:
                return max(Dimension._v(test_value, space) for test_value in test)
            test = value.split("-")
            if len(test) > 1:
                val = Dimension._v(test[0], space)
                for i in range(1, len(test)):
                    val -= Dimension._v(test[i], space)
                return val
            test = value.split("+")
            if len(test) > 1:
                val = Dimension._v(test[0], space)
                for i in range(1, len(test)):
                    val += Dimension._v(test[i], space)
                return val
            if value[-1] == "%":
                return max(int(space * float(value[:-1]) / 100) - 1, 0)
        return int(value)

    def __call__(self, parent=None):
        """
        Evaluates the dimension into a width-height pair.

        Parameters
        ----------
        parent : awsc.termui.block.Block, optional
            The parent dimensions to use for reference. Uses the available space screen if None.

        Returns
        -------
        tuple(int, int)
            A tuple in the form of (width, height) as evaluated values.
        """
        dim = Commons.UIInstance.dim if parent is None else parent.dimensions
        width = Dimension._v(self.width, dim[0])
        height = Dimension._v(self.height, dim[1])
        return (width, height)

    def __str__(self):
        return f"Dim[{self.width}, {self.height}]"


class AbstractAnchor:
    """
    Anchor base class defining the interface anchors must implement.
    """

    def anchor(self, parent):
        """
        Returns the position of the anchor point of the block.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the anchor point.
        """
        return (0, 0)

    def topleft(self, dim, parent):
        """
        Returns the position of the top left corner of the anchored block.

        Parameters
        ----------
        dim : awsc.termui.alignment.Dimension
            The dimensions of the block being anchored.
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        return (0, 0)

    def __str__(self):
        return "@[0, 0]"


class TopLeftAnchor(AbstractAnchor):
    """
    Anchor class which specifies a block is aligned with the top-left corner of the parent at a fixed offset.

    Anchor point: top left corner of the block.

    Parent point: top left corner of the block.

    Attributes
    ----------
    top : int
        The vertical offset from the top left corner.
    left : int
        The horizontal offset from the top left corner.
    """

    def __init__(self, left, top):
        """
        Initializes a TopLeftAnchor object.

        Parameters
        ----------
        left : int
            The horizontal offset from the top left corner.
        top : int
            The vertical offset from the top left corner.
        """
        self.top = top
        self.left = left

    def anchor(self, parent):
        """
        Returns the position of the anchor point of the block.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the anchor point (top left corner).
        """
        if parent is None:
            return (self.left, self.top)
        tlp = parent.topleft()
        return (tlp[0] + self.left, tlp[1] + self.top)

    def topleft(self, dim, parent):
        """
        Returns the position of the top left corner of the anchored block.

        Parameters
        ----------
        dim : awsc.termui.alignment.Dimension
            The dimensions of the block being anchored.
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        return self.anchor(parent)

    def __str__(self):
        return f"@[{self.left}, {self.top}]"


class TopLeftDimensionAnchor(TopLeftAnchor):
    """
    A TopLeft anchor that allows dimension specifications for each axis. If not needed, use TopLeftAnchor for better performance.
    """

    def anchor(self, parent):
        dim = Dimension(self.left, self.top)
        relative = dim(parent)
        parent_anchor = parent.alignment.topleft(parent.dimensions, parent.parent)
        return (parent_anchor[0] + relative[0], parent_anchor[1] + relative[1])

    def __str__(self):
        return f"@Dim[{self.left}, {self.top}]"


class BottomLeftAnchor(AbstractAnchor):
    """
    Anchor class which specifies a block is aligned with the bottom-left corner of the parent at a fixed offset.

    Anchor point: bottom left corner of the block.

    Parent point: bottom left corner of the block.

    Attributes
    ----------
    bottom : int
        The vertical offset from the bottom left corner.
    left : int
        The horizontal offset from the bottom left corner.
    """

    def __init__(self, left, bottom):
        """
        Initializes a BottomLeftAnchor object.

        Parameters
        ----------
        left : int
            The horizontal offset from the bottom left corner.
        bottom : int
            The vertical offset from the bottom left corner.
        """
        self.bottom = bottom
        self.left = left

    def anchor(self, parent):
        """
        Returns the position of the anchor point of the block.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the anchor point (bottom left corner).
        """
        if parent is None:
            return (self.left, self.bottom)
        tlp = parent.topleft()
        brp = parent.bottomright()
        return (tlp[0] + self.left, brp[1] - self.bottom)

    def topleft(self, dim, parent):
        """
        Returns the position of the top left corner of the anchored block.

        Parameters
        ----------
        dim : awsc.termui.alignment.Dimension
            The dimensions of the block being anchored.
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        anchor = self.anchor(parent)
        return (anchor[0], anchor[1] - dim[1] + 1)

    def __str__(self):
        return f"@[{self.left}, {self.bottom}]BL"


class TopRightAnchor(AbstractAnchor):
    """
    Anchor class which specifies a block is aligned with the top-right corner of the parent at a fixed offset.

    Anchor point: top right corner of the block.

    Parent point: top right corner of the block.

    Attributes
    ----------
    top : int
        The vertical offset from the top right corner.
    right : int
        The horizontal offset from the top right corner.
    """

    def __init__(self, right, top):
        """
        Initializes a TopRightAnchor object.

        Parameters
        ----------
        right : int
            The horizontal offset from the top right corner.
        top : int
            The vertical offset from the top right corner.
        """
        self.top = top
        self.right = right

    def anchor(self, parent):
        """
        Returns the position of the anchor point of the block.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the anchor point (top right corner).
        """
        if parent is None:
            return (self.right, self.top)
        tlp = parent.topleft()
        brp = parent.bottomright()
        return (brp[0] - self.right, tlp[1] + self.top)

    def topleft(self, dim, parent):
        """
        Returns the position of the top left corner of the anchored block.

        Parameters
        ----------
        dim : awsc.termui.alignment.Dimension
            The dimensions of the block being anchored.
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        anchor = self.anchor(parent)
        return (anchor[0] - dim[0] + 1, anchor[1])

    def __str__(self):
        return f"@[{self.right}, {self.top}]TR"


class CenterAnchor(AbstractAnchor):
    """
    Anchor class which specifies a block is aligned with the center of the parent at a fixed offset.

    Anchor point: center of the block.

    Parent point: center of the block.

    Attributes
    ----------
    xoffset : int
        The horizontal offset from the center.
    yoffset : int
        The vertical offset from the center
    """

    def __init__(self, xoffset, yoffset):
        """
        Initializes a CenterAnchor object.

        Parameters
        ----------
        xoffset : int
            The horizontal offset from the center.
        yoffset : int
            The vertical offset from the center
        """
        self.xoffset = xoffset
        self.yoffset = yoffset

    def anchor(self, parent):
        """
        Returns the position of the anchor point of the block.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the anchor point (center).
        """
        tlp = (0, 0)
        dim = Commons.UIInstance.dim
        if parent is not None:
            tlp = parent.topleft()
            dim = parent.dimensions()
        return (
            int(dim[0] / 2) + self.xoffset + tlp[0],
            int(dim[1] / 2) + self.yoffset + tlp[1],
        )

    def topleft(self, dim, parent):
        """
        Returns the position of the top left corner of the anchored block.

        Parameters
        ----------
        dim : awsc.termui.alignment.Dimension
            The dimensions of the block being anchored.
        parent : awsc.termui.block.Block
            The parent block to anchor to. If None, the entire screen is considered the parent block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        anchor = self.anchor(parent)
        return (anchor[0] - int(dim[0] / 2), anchor[1] - int(dim[1] / 2))

    def __str__(self):
        return f"@[{self.xoffset}, {self.yoffset}]C"



"""
Module for security group resources.
"""
from .base_control import Describer, MultiLister, ResourceLister
from .common import Common
from .resource_common import multilister_with_compare_path


class SGDescriber(Describer):
    """
    Describer control for security group resources.
    """

    prefix = "sg_browser"
    title = "Security Group"

    resource_type = "security group"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Security Group"
    describe_method = "describe_security_groups"
    describe_kwarg_name = "GroupIds"
    describe_kwarg_is_list = True
    object_path = ".SecurityGroups[0]"
    default_entry_key = "group id"


class SGRelated(MultiLister):
    """
    Lister control for AWS resources using a certain security group.
    """

    prefix = "sg_related"
    title = "Resources using Security Group"

    def title_info(self):
        return self.compare_value

    def __init__(self, *args, **kwargs):
        kwargs["compare_key"] = "group id"
        self.resource_descriptors = [
            multilister_with_compare_path("ec2", "[.SecurityGroups[].GroupId]", True),
            multilister_with_compare_path(
                "rds", "[.VpcSecurityGroups[].VpcSecurityGroupId]", True
            ),
        ]
        super().__init__(*args, **kwargs)


def _sg_determine_ingress_rules(self, result):
    return len(result["IpPermissions"])


def _sg_determine_egress_rules(self, result):
    return len(result["IpPermissionsEgress"])


class SGResourceLister(ResourceLister):
    """
    Lister control for security group resources.
    """

    prefix = "sg_list"
    title = "Security Groups"
    command_palette = ["sg", "securitygroup"]

    resource_type = "security group"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Security Group"
    list_method = "describe_security_groups"
    item_path = ".SecurityGroups"
    columns = {
        "group id": {
            "path": ".GroupId",
            "size": 30,
            "weight": 0,
        },
        "name": {
            "path": ".GroupName",
            "size": 30,
            "weight": 1,
            "sort_weight": 0,
        },
        "vpc": {
            "path": ".VpcId",
            "size": 15,
            "weight": 2,
        },
        "ingress rules": {
            "path": _sg_determine_ingress_rules,
            "size": 15,
            "weight": 3,
        },
        "egress rules": {
            "path": _sg_determine_egress_rules,
            "size": 15,
            "weight": 4,
        },
    }
    describe_command = SGDescriber.opener
    open_command = SGRelated.opener
    open_selection_arg = "compare_value"
    primary_key = "group id"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @ResourceLister.Autohotkey("i", "View ingress rules", True)
    def ingress(self, *args):
        """
        Hotkey callback for viewing ingress rules.
        """
        if self.selection is not None:
            Common.Session.push_frame(SGRuleLister.opener(sg_entry=self.selection))

    @ResourceLister.Autohotkey("e", "View egress rules", True)
    def egress(self, *args):
        """
        Hotkey callback for viewing egress rules.
        """
        if self.selection is not None:
            Common.Session.push_frame(
                SGRuleLister.opener(sg_entry=self.selection, egress=True)
            )


_sg_well_known = {
    20: "ftp",
    22: "ssh",
    23: "telnet",
    25: "smtp",
    80: "http",
    110: "pop3",
    220: "imap",
    443: "https",
    989: "ftps",
    1437: "dashboard-agent",
    1443: "mssql",
    3306: "mysql",
    3389: "rdp",
    5432: "pgsql",
    6379: "redis",
    8983: "solr",
    9200: "es",
}


def _sg_rule_determine_name(self, rule):
    if rule["IpProtocol"] == "-1":
        return "<all>"
    if rule["FromPort"] != rule["ToPort"]:
        return ""
    if rule["FromPort"] not in _sg_well_known:
        return ""
    return _sg_well_known[rule["FromPort"]]


def _sg_rule_determine_protocol(self, rule):
    if rule["IpProtocol"] == "-1" or rule["IpProtocol"] == -1:
        return "<all>"
    return rule["IpProtocol"]


def _sg_rule_determine_ips(self, rule):
    if "IpRanges" not in rule:
        return ""
    return ",".join(cidrs["CidrIp"] for cidrs in rule["IpRanges"])


def _sg_rule_determine_sgs(self, rule):
    if "UserIdGroupPairs" not in rule:
        return ""
    return ",".join(pair["GroupId"] for pair in rule["UserIdGroupPairs"])


def _sg_rule_determine_port_range(self, rule):
    if rule["IpProtocol"] == "-1":
        return "0-65535"
    if rule["FromPort"] != rule["ToPort"]:
        return f"{rule['FromPort']}-{rule['ToPort']}"
    return str(rule["FromPort"])


class SGRuleLister(ResourceLister):
    """
    Lister object for security group rules.
    """

    prefix = "sg_rule_list"
    title = "SG Rules"

    resource_type = "security group"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Security Group"
    list_method = "describe_security_groups"
    item_path = ".SecurityGroups"
    columns = {
        "protocol": {
            "path": _sg_rule_determine_protocol,
            "size": 10,
            "weight": 0,
        },
        "name": {
            "path": _sg_rule_determine_name,
            "size": 5,
            "weight": 1,
        },
        "ip addresses": {
            "path": _sg_rule_determine_ips,
            "size": 40,
            "weight": 2,
        },
        "security groups": {
            "path": _sg_rule_determine_sgs,
            "size": 50,
            "weight": 3,
        },
        "port range": {
            "path": _sg_rule_determine_port_range,
            "size": 20,
            "weight": 4,
            "sort_weight": 0,
        },
    }
    describe_command = SGDescriber.opener
    open_command = SGRelated.opener
    open_selection_arg = "compare_value"
    primary_key = "group id"

    def title_info(self):
        gress = "Egress" if self.egress else "Ingress"
        return f"{gress}: {self.sg_entry['group id']}"

    def __init__(
        self,
        *args,
        sg_entry=None,
        egress=False,
        **kwargs,
    ):
        if sg_entry is None:
            raise ValueError("sg_entry is required")
        self.sg_entry = sg_entry
        self.egress = egress
        self.list_kwargs = {"GroupIds": [self.sg_entry["group id"]]}
        gress = "Egress" if egress else ""
        self.item_path = f".SecurityGroups[0].IpPermissions{gress}"
        super().__init__(*args, **kwargs)



"""
This module defines the UI handler class, which is where all the magic happens.
"""
import shutil
import signal
import tempfile
import termios
import threading
import time
import tty
from pathlib import Path

import chardet
import magic
from blessed import Terminal

from .alignment import Dimension, TopLeftAnchor
from .block import Block
from .common import Commons
from .control import Control
from .screen import Screen

FRAMERATE = 0.02
"""
Defines at most how often should a new frame be generated, in seconds. This is a terminal application, we really don't need more than 50 FPS.
"""


class ControlCodes:
    """
    Enum for available CTRL codes. See the documentation of awsc.termui.control.HotkeyControl.add_hotkey on how to use these.
    """

    A = "\x01"
    B = "\x02"
    C = "\x03"
    D = "\x04"
    E = "\x05"
    F = "\x06"
    G = "\x07"
    K = "\x0b"
    L = "\x0c"
    M = "\x0d"
    N = "\x0e"
    O = "\x0f"
    P = "\x10"
    Q = "\x11"
    R = "\x12"
    S = "\x13"
    T = "\x14"
    U = "\x15"
    V = "\x16"
    W = "\x17"
    X = "\x18"
    Y = "\x19"
    Z = "\x1a"


class KeyDisplay(Control):
    """
    Debug control for displaying the key being pressed. Input function of this control does not consume the key.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes a KeyDisplay object.
        """
        super().__init__(*args, **kwargs)
        self.text = "<no key>"

    def input(self, key):
        if key.is_sequence:
            self.text = f"{(str(key), key.name, key.code)}"
        else:
            self.text = key
        return False

    def paint(self):
        super().paint()
        topleft = self.topleft()
        Commons.UIInstance.print(self.text, xy=topleft)


class UI:
    """
    UI singleton object. Handles the main loop of a termui application.

    Attributes
    ----------
    debug : int
        Debug level. Should be unused most of the time.
    top_block : awsc.termui.block.Block
        The topmost block which spans the screen. Should be used as the parent of all top level blocks displayed on the screen.
    last_paint : time.Time
        Timer tracking when the last paint event occurred.
    last_size_refresh : time.Time
        Timer tracking when the last refresh of screen size occurred. Screen size refresh is an expensive operation for some reason.
    term : blessed.Terminal
        The blessed Terminal that is being operated on.
    _w : int
        The width of the terminal in characters. Eventually consistent value, only changes on screen size refresh.
    _h : int
        The height of the terminal in characters. Eventually consistent value, only changes on screen size refresh.
    exit : bool
        If set, the program will exit on the next cycle of the main loop.
    buf : awsc.termui.screen.Screen
        The screen buffer that is manipulated in paint operations. Each control outputting directly to the screen is expensive and
        error prone. Manipulating a buffer and printing it all at once is significantly faster.
    tickers : list(callable)
        A list of hooks to execute on each main loop cycle.
    cache_dir : str
        A file cache temporary directory. Ideally erased upon graceful shutdown, with all of its contents.
    mutex : threading.Lock
        Controls access to the input buffer.
    input_buffer : list(blessed.keyboard.Keystroke)
        A list of unprocessed keyboard inputs. Handled in the event loop.
    dirty : bool
        Whether the UI requires a repaint.
    restore : object
        Stores the state of the terminal to restore to upon shutdown. Required to return the user to a working terminal after exiting.
    suspend_input_buffer : bool
        Atomic for disabling input buffering during unraw operations.
    """

    def __init__(self):
        """
        Initializes a UI object.
        """
        if Commons.UIInstance is not None:
            raise RuntimeError("UI is a singleton.")
        Commons.UIInstance = self
        self.debug = 1
        self.top_block = Block(
            None, TopLeftAnchor(0, 0), Dimension("100%", "100%"), tag="top"
        )
        self.last_paint = time.time()
        self.last_size_refresh = time.time()
        self.term = Terminal()
        self._w = self.term.width
        self._h = self.term.height
        self.exit = False
        signal.signal(signal.SIGINT, self.set_exit)
        self.buf = Screen(self)
        self.tickers = []
        self.cache_dir = None
        self.mutex = threading.Lock()
        self.input_buffer = []
        self.dirty = False
        self.restore = None
        self.suspend_input_buffer = False

    @staticmethod
    def read_file(filename):
        """
        Reads a file in the correct encoding, or binary mode if mime type does not indicate it is a text file.

        Parameters
        ----------
        filename : str or pathlib.Path
            The name of the file.

        Returns
        -------
        str or bytes
            The contents of the file.
        """
        mime = magic.from_file(filename)
        filename = str(filename)
        with open(filename, "rb") as file:
            data = file.read()
        # Do not even try to decode common binary format mime types.
        if not mime.startswith("text/"):
            return data
        result = chardet.detect(data)
        # Arbitrary lower limit for confidence.
        if result["confidence"] >= 0.7:
            return data.decode(result["encoding"])
        return data

    def filecache(self, obj, callback, *args, **kwargs):
        """
        Retrieves a cached file from the temporary file cache. If the file is not present, the callback is called to retrieve the cached file.

        Parameters
        ----------
        obj : str
            The name of the cached file to retrieve.
        callback : callable(str, *args, **kwargs) -> str or bytes
            A callback that is used to fetch the file contents if it is not cached. The callback, for now, is expected to cache the file
            itself so that it can be fetched from cache later. The callback is passed the full path where the file should be cached, as well
            as all additional positional and keyword arguments this function receives and should return the contents of the file as str or
            bytes, depending on whether it's a text or binary file.
        *args : list
            A list of additional positional arguments to pass to the callback.
        **kwargs : dict
            A dict of additional keyword arguments to pass to the callback.

        Returns
        -------
        str or bytes
            The contents of the cached file.
        """
        objpath = self.filecache_path(obj, callback, *args, **kwargs)
        return self.read_file(objpath)

    def filecache_path(self, obj, callback, *args, **kwargs):
        """
        Like filecache, except this one only returns the path of the cached file. Useful if you don't want to decode and recode the file.

         Parameters
        ----------
        obj : str
            The name of the cached file to retrieve.
        callback : callable(str, *args, **kwargs) -> str or bytes
            A callback that is used to fetch the file contents if it is not cached. The callback, for now, is expected to cache the file
            itself so that it can be fetched from cache later. The callback is passed the full path where the file should be cached, as well
            as all additional positional and keyword arguments this function receives and should return the contents of the file as str or
            bytes, depending on whether it's a text or binary file.
        *args : list
            A list of additional positional arguments to pass to the callback.
        **kwargs : dict
            A dict of additional keyword arguments to pass to the callback.

        Returns
        -------
        pathlib.Path
            The path to the file.
        """
        if self.cache_dir is None:
            self.cache_dir = tempfile.mkdtemp(prefix="awsc")
        objpath = Path(self.cache_dir) / obj
        if not objpath.exists():
            callback(str(objpath.resolve()), *args, **kwargs)
        return objpath

    def print_centered(self, out, bounds, color=None, bold=False):
        """
        Naive center-printing function. Does not currently handle line breaks, so the text must fit into one line of the boundary.

        Parameters
        ----------
        out : str
            The string to print on the screen.
        bounds : tuple(tuple(int, int), tuple(int, int))
            The bounding box for printing the text.
        color : awsc.termui.color.Color, optional
            A callable which takes a string and returns a string. Used as a callback to color the output string.
        bold : bool, default=False
            If set, prints the text in bold.
        """
        center_row = int((bounds[1][1] - bounds[1][0]) / 2) - 2
        center_col = int((bounds[0][1] - bounds[0][0]) / 2)
        start_col = int(center_col - (len(out) / 2))
        self.print(
            out,
            xy=(bounds[0][0] + start_col, bounds[1][0] + center_row),
            color=color,
            bounds=bounds,
            bold=bold,
        )

    def print(self, out, xy=None, color=None, wrap=False, bounds=None, bold=False):
        """
        Prints a text to a specified area of the screen buffer. Has no real effect if called outside the block paint hooks.

        Parameters
        ----------
        out : str
            The string to print on the screen.
        xy : tuple(int, int), optional
            Where to print on the screen. If omitted, prints to the top left corner of the bounding box.
        color : awsc.termui.color.Color, optional
            A callable which takes a string and returns a string. Used as a callback to color the output string.
        wrap : bool, default=False
            If set, text will wrap within the bounding box. Otherwise, the overflow on the row of the text is cut off.
        bounds : tuple(tuple(int, int), tuple(int, int)), optional
            The bounding box for printing the text. If omitted, the bounding box is the entire terminal.
        bold : bool, default=False
            If set, prints the text in bold.
        """
        if color is not None and not callable(color):
            raise ValueError("Color must be callable or None.")
        self.dirty = True
        if bounds is None:
            bounds = ((0, self.width), (0, self.height))
        if xy is None:
            xy = (bounds[0][0], bounds[1][0])
        end = False
        if (xy[0] > bounds[0][1] and not wrap) or xy[1] > bounds[1][1]:
            return
        while not end:
            space = bounds[0][1] - xy[0]
            if len(out) > space:
                part = out[:space]
                out = out[space:]
            else:
                space = len(out)
                part = out
                out = ""
                end = True
            position = 0
            for i in range(xy[0], xy[0] + space):
                try:
                    char = self.buf[xy[1]][i]
                    char.value = part[position]
                    char.color = color
                    char.bold = bold
                    char.dirty = True
                except IndexError:
                    break
                position += 1
            xy = (bounds[0][0], xy[1] + 1)
            if not wrap or xy[1] >= bounds[1][1]:
                end = True

    def refresh_size(self):
        """
        Refreshes the cached width and height of the terminal if it has not been refreshed recently.

        Based on profiling, retrieving the blessed terminal width and height is a rather expensive operation, which is why this is
        strictly rate limited and cached. Besides, the user is very unlikely to repeatedly and rapidly resize the terminal.
        """
        if self.last_paint - self.last_size_refresh > 1:
            self.last_size_refresh = self.last_paint
            self._w = self.term.width
            self._h = self.term.height

    @property
    def width(self):
        """
        Read-only property for the width of the terminal. Refreshes the value of _w and _h if necessary.

        Returns
        -------
        int
            The width of the terminal.
        """
        self.refresh_size()
        return self._w

    @property
    def height(self):
        """
        Read-only property for the height of the terminal. Refreshes the value of _w and _h if necessary.

        Returns
        -------
        int
            The height of the terminal.
        """
        self.refresh_size()
        return self._h

    @property
    def dim(self):
        """
        Read-only property for the dimensions of the terminal. Refreshes the value of _w and _h if necessary.

        Returns
        -------
        tuple(int, int)
            The width and height of the terminal.
        """
        self.refresh_size()
        return (self._w, self._h)

    def set_exit(self, *args, **kwargs):
        """
        Hotkey-compatible function that sets the graceful shutdown flag on the UI.

        Parameters
        ----------
        *args : list
            Unused.
        **kwargs : dict
            Unused.
        """
        self.exit = True

    def before_paint(self):
        """
        Pre-paint hook for the UI. Executes the pre-paint hooks of all blocks within the UI.
        """
        self.top_block.before_paint()

    def progress_bar_paint(self, perc):
        """
        Convenience function for painting a progress bar. The progress bar is painted on the bottom row of the terminal.

        Has no effect if called outside of the paint hooks of blocks.s

        Parameters
        ----------
        perc : float
            The percentage of the progress bar to fill. Should be between 0 and 1, inclusive, for sanity's sake.
        """
        width = int(self.width * perc)
        with self.term.location(0, self.height - 1):
            percentage = perc * 100
            perc_t = f"{percentage:.2f}% "
            print(
                "\033[38;5;220m" + perc_t + ((width - len(perc_t)) * "░") + "\033[0m",
                end="",
            )

    def check_one_key(self):
        """
        Retrieves a single keystroke within a timeout half a frame. The polling time is intentionally short so the keyboard event hook does not
        delay the main loop by a significant degree if no keys are being pressed.

        Returns
        -------
        blessed.keyboard.Keystroke
            A keystroke, if a keystroke happened, or None if no keystrokes happened.
        """
        return self.term.inkey(FRAMERATE / 2.0, FRAMERATE / 2.0)

    def paint(self):
        """
        Main paint handler of the UI. Clears the screen buffer and prompts all blocks within the UI to be painted.
        """
        self.buf.clear()
        self.top_block.paint()
        with self.term.location(0, 0):
            self.buf.output()
        self.last_paint = time.time()
        self.dirty = False

    # TODO: Check BUG documented in the doc.
    def unraw(self, cmd, *args, **kwargs):
        """
        Runs a callback in a default terminal setting. Required to run external commands such as nano or vi in a way that the user
        would expect.

        KNOWN BUG: For some reason, a single keypress erroneously has its control character (^) consumed but the rest of the sequence is
        output if a key is pressed roughly 1 second after the callback is executed.
        It's annoying but not breaking anything, but does require attention at some point.

        Parameters
        ----------
        cmd : callable
            The function to call after resetting the terminal settings.
        *args : list
            A list of positional arguments to pass to the command.
        **kwargs : dict
            A map of keyword arguments to pass to the command.
        """
        # pylint: disable=protected-access
        termios.tcflush(self.term._keyboard_fd, termios.TCIOFLUSH)
        # pylint: disable=protected-access
        termios.tcsetattr(self.term._keyboard_fd, termios.TCSAFLUSH, self.restore)
        # pylint: disable=protected-access
        # termios.tcsetattr(self.term._keyboard_fd, termios.TCSANOW, self.restore)
        # pylint: disable=protected-access
        self.term._line_buffered = True
        self.term.stream.write(self.term.normal_cursor)
        self.term.stream.flush()
        self.term.stream.write(self.term.exit_fullscreen)
        self.term.stream.flush()
        self.suspend_input_buffer = True
        time.sleep(FRAMERATE * 10)
        # pylint: disable=protected-access
        termios.tcflush(self.term._keyboard_fd, termios.TCIOFLUSH)
        try:
            return cmd(*args, **kwargs)
        finally:
            self.term.stream.write(self.term.enter_fullscreen)
            self.term.stream.flush()
            self.term.stream.write(self.term.hide_cursor)
            self.term.stream.flush()
            self.suspend_input_buffer = False
            # pylint: disable=protected-access
            tty.setraw(self.term._keyboard_fd, termios.TCSANOW)
            # pylint: disable=protected-access
            self.term._line_buffered = False

    def buffer_input(self):
        """
        Input buffer loop. Runs on a separate thread, listens for keystrokes and pushes them into the input buffer as they become available.

        Actual handling of the inputs happens on the main loop thread.
        """
        while True:
            if self.suspend_input_buffer:
                time.sleep(FRAMERATE * 10)
                continue
            key = self.term.inkey(None, FRAMERATE / 2)
            if key is not None and key != "":
                with self.mutex:
                    self.input_buffer.append(key)

    def process_input_buffer(self, start_time):
        """
        Processes part of the input buffer. The input buffer will not process any inputs within the final 10ms of its allowed frame time.

        Parameters
        ----------
        start_time: time.Time
            The exact time when the frame processing started.
        """
        with self.mutex:
            if len(self.input_buffer) > 0:
                key = self.input_buffer[0]
                self.input_buffer = self.input_buffer[1:]
                if key == "\x03":
                    raise KeyboardInterrupt
                self.top_block.input(key)
                curr_time = time.time()
                if curr_time - start_time > FRAMERATE - 0.02:
                    return

    def main(self):
        """
        Main loop function of the UI. Call this after setting up your application.

        The main loop function will configure the terminal to use raw inputs, start up the keyboard buffer loop, then runs until the graceful
        shutdown flag is set in an infinite loop. Within this loop, in each cycle, first, the input buffer is processed, then a new frame is
        rendered.
        """
        # pylint: disable=protected-access
        self.restore = termios.tcgetattr(self.term._keyboard_fd)
        try:
            with self.term.fullscreen():
                with self.term.hidden_cursor():
                    with self.term.raw():
                        thread = threading.Thread(target=self.buffer_input, daemon=True)
                        thread.start()
                        while not self.exit:
                            # print(self.term.clear())
                            start_time = time.time()
                            self.process_input_buffer(start_time)
                            for func in self.tickers:
                                func()
                            self.before_paint()
                            if self.dirty or time.time() - self.last_paint > 3:
                                self.paint()
                            curr_time = time.time()
                            delay = FRAMERATE - (curr_time - start_time)
                            if delay > 0:
                                time.sleep(delay)
        except KeyboardInterrupt:
            pass
        finally:
            if self.cache_dir is not None:
                shutil.rmtree(self.cache_dir)



"""
Module which contains AWS context-related resources.

Contexts are keypairs able to access different AWS accounts.
"""
import configparser
import datetime
import sys
from pathlib import Path

from botocore import exceptions as botoerror

from .base_control import (
    DeleteResourceDialog,
    DialogFieldResourceListSelector,
    OpenableListControl,
)
from .common import Common, SessionAwareDialog
from .termui.control import Border
from .termui.dialog import DialogFieldLabel, DialogFieldText
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes


class ImportContextDialog(SessionAwareDialog):
    """
    Dialog control for importing a context from the environment.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    name_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the context.
    access_key : str
        The access key being imported.
    secret_key : str
        The secret key being imported.
    access_key_field : awsc.termui.dialog.DialogFieldLabel
        Field for displaying the access key being imported.
    secret_key_field : awsc.termui.dialog.DialogFieldLabel
        Field for displaying that a secret key is being imported. It is masked.
    """

    line_size = 20

    def __init__(self, parent, alignment, dimensions, *args, caller=None, **kwargs):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "New Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.caller = caller
        self.set_title_label("Enter AWS context details")
        self.name_field = DialogFieldText(
            "Name:", **Common.textfield_colors("context_add")
        )
        self.add_field(self.name_field)
        creds = Common.Session.service_provider.env_session().get_credentials()
        if creds is None or not creds.access_key or not creds.secret_key:
            Common.Session.set_message(
                "No valid credentials in environment.", Common.color("message_error")
            )
            self.close()
            return
        self.access_key = creds.access_key
        self.secret_key = creds.secret_key
        self.access_key_field = DialogFieldLabel(
            [
                ("Access key: ", Common.color("generic")),
                (creds.access_key, Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldLabel(
            [
                ("Secret key: ", Common.color("generic")),
                ("*" * len(creds.secret_key), Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.secret_key_field)

    def input(self, key):
        if not self.accepts_inputs:
            return True
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={"access": self.access_key, "secret": self.secret_key}
            )
        except botoerror.ClientError as error:
            self.error_label.text = "Key verification failed"
            Common.clienterror(
                error,
                "Key verification failed",
                "Core",
                subcategory="STS",
                resource=self.access_key,
                set_message=False,
                access_key=self.access_key,
                secret_key="<PRESENT>",
            )
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text, sts["Account"], self.access_key, self.secret_key
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


def _context_color_defaults():
    return {
        "color": Common.color(
            "context_add_modal_dialog_textfield", "modal_dialog_textfield"
        ),
        "selected_color": Common.color(
            "context_add_modal_dialog_textfield_selected",
            "modal_dialog_textfield_selected",
        ),
        "label_color": Common.color(
            "context_add_modal_dialog_textfield_label",
            "modal_dialog_textfield_label",
        ),
    }


class MFADialog(SessionAwareDialog):
    """
    Dialog control for multi-factor authentication.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    token_field : aws.termui.dialog.DialogFieldText
        Field for entering the MFA token.
    """

    line_size = 20

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        caller=None,
        **kwargs,
    ):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Authenticate context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.set_title_label(
            f"Enter MFA token code for {Common.Session.context_data['mfa_device']}"
        )
        self.token_field = DialogFieldText(
            "MFA Token:", label_min=16, **_context_color_defaults()
        )
        self.add_field(self.token_field)

    def accept_and_close(self):
        if self.token_field.text == "":
            self.error_label.text = "Token cannot be blank."
            return
        keys = Common.Session.context_perm_auth
        resp = Common.generic_api_call(
            "sts",
            "get_session_token",
            {
                "DurationSeconds": 86400,
                "SerialNumber": Common.Session.context_data["mfa_device"],
                "TokenCode": self.token_field.text,
            },
            "Retrieve session token",
            "STS",
            keys=keys,
        )
        if not resp["Success"]:
            self.error_label.text = "Failed to acquire security credentials."
            return
        creds = resp["Response"]["Credentials"]
        Common.Configuration.keystore.set_temp(
            Common.Session.context,
            {
                "access": creds["AccessKeyId"],
                "secret": creds["SecretAccessKey"],
                "session": creds["SessionToken"],
                "expiry": creds["Expiration"].timestamp(),
                "mfa_device": Common.Session.context_data["mfa_device"],
            },
        )

        self.accepts_inputs = False
        Common.Session.context = Common.Session.context
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class AddContextDialog(SessionAwareDialog):
    """
    Dialog control for adding a new context.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    name_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the context.
    access_key_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the access key.
    secret_key_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the secret key.
    """

    line_size = 20

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        caller=None,
        existing_context=None,
        **kwargs,
    ):
        from .resource_iam_user import MFADeviceLister

        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            f"{'New' if existing_context is None else 'Edit'} Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.editing = existing_context
        self.set_title_label("Enter AWS context details")
        self.name_field = (
            DialogFieldText(
                "Name:",
                label_min=16,
                **_context_color_defaults(),
            )
            if existing_context is None
            else DialogFieldLabel(
                [
                    ("Name: ", Common.color("generic")),
                    (existing_context, Common.color("highlight")),
                ],
                centered=False,
            )
        )
        keys = (
            {"access": "", "secret": ""}
            if existing_context is None
            else Common.Configuration.keystore.get_permanent_credentials(
                existing_context
            )
        )
        self.add_field(self.name_field)
        self.access_key_field = DialogFieldText(
            "Access key:",
            text=keys["access"],
            label_min=16,
            **_context_color_defaults(),
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldText(
            "Secret key:",
            text=keys["secret"],
            label_min=16,
            **_context_color_defaults(),
            password=True,
        )
        self.add_field(self.secret_key_field)
        mfa_device = (
            ""
            if existing_context is None
            else Common.Configuration["contexts"][existing_context]["mfa_device"]
        )
        self.mfa_device_field = DialogFieldResourceListSelector(
            MFADeviceLister,
            "MFA device:",
            default=mfa_device,
            label_min=16,
            **_context_color_defaults(),
        )
        self.add_field(self.mfa_device_field)
        self.caller = caller

    def input(self, key):
        if not self.accepts_inputs:
            return True
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        if self.editing is None:
            if self.name_field.text == "":
                self.error_label.text = "Name cannot be blank."
                return
            if self.name_field.text == "localstack":
                self.error_label.text = (
                    'The name "localstack" is protected and cannot be used.'
                )
                return
        if self.access_key_field.text == "":
            self.error_label.text = "Access key cannot be blank."
            return
        if self.secret_key_field.text == "":
            self.error_label.text = "Secret key cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={
                    "access": self.access_key_field.text,
                    "secret": self.secret_key_field.text,
                }
            )
        except botoerror.ClientError as error:
            self.error_label.text = "Key verification failed"
            Common.clienterror(
                error,
                "Key verification failed",
                "Core",
                subcategory="STS",
                resource=self.access_key_field.text,
                set_message=False,
                access_key=self.access_key_field.text,
                secret_key="<PRESENT>",
            )
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text if self.editing is None else self.editing,
            sts["Account"],
            self.access_key_field.text,
            self.secret_key_field.text,
            self.mfa_device_field.text,
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class AddRoleContextDialog(SessionAwareDialog):
    """
    Dialog control for adding a new context.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    name_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the context.
    source_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the source context.
    role_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the role to assume.
    """

    line_size = 20

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        caller=None,
        existing_context=None,
        **kwargs,
    ):
        from .resource_iam_role import RoleLister

        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            f"{'New' if existing_context is None else 'Edit'} Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.editing = existing_context
        self.set_title_label("Enter AWS context details")
        self.name_field = (
            DialogFieldText("Name:", label_min=16, **_context_color_defaults())
            if existing_context is None
            else DialogFieldLabel(
                [
                    ("Name: ", Common.color("generic")),
                    (existing_context, Common.color("highlight")),
                ],
                centered=False,
            )
        )
        self.add_field(self.name_field)
        source = (
            ""
            if existing_context is None
            else Common.Configuration.keystore.get_ref_name(existing_context)
        )
        self.source_field = DialogFieldResourceListSelector(
            ContextList,
            "Source context:",
            default=source,
            label_min=16,
            **_context_color_defaults(),
        )
        self.add_field(self.source_field)

        role = (
            ""
            if existing_context is None
            else Common.Configuration["contexts"][existing_context]["role"]
        )
        self.role_field = DialogFieldResourceListSelector(
            RoleLister,
            "IAM role to assume:",
            default=role,
            label_min=16,
            **_context_color_defaults(),
        )
        self.add_field(self.role_field)
        self.caller = caller

    def input(self, key):
        if not self.accepts_inputs:
            return True
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        if self.editing is None:
            if self.name_field.text == "":
                self.error_label.text = "Name cannot be blank."
                return
            if self.name_field.text == "localstack":
                self.error_label.text = (
                    'The name "localstack" is protected and cannot be used.'
                )
                return
        if self.source_field.text == "":
            self.error_label.text = "Source context cannot be blank."
            return
        if self.role_field.text == "":
            self.error_label.text = "IAM Role cannot be blank."
            return

        ctx = Common.Configuration.keystore.get_permanent_credentials(
            self.source_field.text
        )
        sts = Common.Session.service_provider.whoami(keys=ctx)

        Common.Configuration.add_or_edit_role_context(
            self.name_field.text if self.editing is None else self.editing,
            sts["Account"],
            self.source_field.text,
            self.role_field.text,
            mfa_device="",
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class ContextList(OpenableListControl):
    """
    Lister control for contexts.
    """

    prefix = "context_list"
    title = "Contexts"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_column("account id", 12)
        self.add_column("default", 7)
        self.add_column("type", 4)

        self.jump_cursor = True
        self.reload_contexts()
        self.jump_cursor = False

        if self.selector_cb is None:
            if Common.Session.context != "":
                self.select_context(self, initial=True)

    def reload_contexts(self):
        """
        Refreshes the list of contexts from configuration.
        """
        self.entries = []
        idx = 0
        for context, data in Common.Configuration["contexts"].items():
            self.add_entry(
                ListEntry(
                    context,
                    **{
                        "account id": data["account_id"],
                        "default": "✓"
                        if context == Common.Configuration["default_context"]
                        else " ",
                        "type": "key pair" if data["role"] == "" else "role",
                    },
                )
            )
            if self.jump_cursor and context == Common.Configuration["default_context"]:
                self.selected = idx
            idx += 1
        if 0 <= self.selected < len(self.entries):
            return
        self.selected = 0

    @OpenableListControl.Autohotkey("a", "Add key context")
    def add_new_context(self, _):
        """
        Hotkey callback for adding a new context.
        """
        AddContextDialog.opener(caller=self)

    @OpenableListControl.Autohotkey("s", "Add role context")
    def add_role_context(self, _):
        """
        Hotkey callback for adding a new role context.
        """
        AddRoleContextDialog.opener(caller=self)

    @OpenableListControl.Autohotkey("e", "Edit context", True)
    def edit_context(self, _):
        """
        Hotkey callback for adding a new context.
        """
        if self.selection["name"] == "localstack":
            Common.Session.set_message(
                "Cannot edit localstack context", Common.color("message_error")
            )
            return
        if self.selection["type"] == "key pair":
            AddContextDialog.opener(
                caller=self, existing_context=self.selection["name"]
            )
        elif self.selection["type"] == "role":
            AddRoleContextDialog.opener(
                caller=self, existing_context=self.selection["name"]
            )

    @OpenableListControl.Autohotkey("x", "Export context", True)
    def export_context(self, _):
        """
        Hotkey callback for exporting a context to .aws credentials. Performs MFA if necessary to export an authenticated context.
        """
        mfa_device = Common.Configuration["contexts"][self.selection["name"]][
            "mfa_device"
        ]
        extra_fields = {
            "label_overwrite": DialogFieldLabel(
                "This will overwrite the AWS profile with the same name.",
                Common.color("modal_dialog_error"),
            )
        }
        if mfa_device != "":
            extra_fields["mfa_device"] = DialogFieldLabel(
                f"Enter a code from MFA device {mfa_device}",
                Common.color("modal_dialog_textfield_label"),
            )
            extra_fields["mfa_code"] = DialogFieldText(
                "MFA code: ", label_min=16, **_context_color_defaults()
            )
        DeleteResourceDialog.opener(
            caller=self,
            resource_type="context",
            resource_identifier=self.selection["name"],
            action_name="export",
            extra_fields=extra_fields,
            callback=self.do_export,
        )

    def do_export(self, **kwargs):
        """
        Callback for confirmation of context export.
        """
        aws_creds = Path.home() / ".aws" / "credentials"
        try:
            with aws_creds.open("r", encoding="utf-8") as file:
                creds = file.read()
        except OSError as error:
            if error.errno == 2:  # FNF
                creds = ""
            else:
                print(
                    f"Failed to open ~/.aws/credentials: {str(error)}",
                    file=sys.stderr,
                )
                return
        keys = Common.Configuration.keystore.get_permanent_credentials(
            self.selection["name"]
        )
        parser = configparser.ConfigParser(default_section="__default")
        parser.read_string(creds)
        if parser.has_section(self.selection["name"]):
            parser.remove_section(self.selection["name"])
        parser.add_section(self.selection["name"])
        mfa_device = Common.Configuration["contexts"][self.selection["name"]][
            "mfa_device"
        ]
        if mfa_device != "":
            parser.set(self.selection["name"], "aws_mfa_device", mfa_device)
            resp = Common.generic_api_call(
                "sts",
                "get_session_token",
                {
                    "DurationSeconds": 86400,
                    "SerialNumber": mfa_device,
                    "TokenCode": kwargs["mfa_code"].value,
                },
                "Retrieve session token",
                "STS",
                keys=keys,
            )
            if not resp["Success"]:
                Common.Session.set_message(
                    "Error fetching security credentials", Common.color("message_error")
                )
                return
            token = resp["Response"]["Credentials"]["SessionToken"]
            expiration = resp["Response"]["Credentials"]["Expiration"]
            parser.set(self.selection["name"], "aws_session_token", token)
            parser.set(self.selection["name"], "aws_security_token", token)
            parser.set(
                self.selection["name"],
                "expiration",
                expiration.strftime("%Y-%m-%d %H:%M:%S"),
            )
            parser.set(
                self.selection["name"],
                "aws_access_key_id",
                resp["Response"]["Credentials"]["AccessKeyId"],
            )
            parser.set(
                self.selection["name"],
                "aws_secret_access_key",
                resp["Response"]["Credentials"]["SecretAccessKey"],
            )
        else:
            parser.set(self.selection["name"], "aws_access_key_id", keys["access"])
            parser.set(self.selection["name"], "aws_secret_access_key", keys["secret"])
        with aws_creds.open("w", encoding="utf-8") as file:
            parser.write(file)

    @OpenableListControl.Autohotkey("i", "Import context")
    def import_context(self, _):
        """
        Hotkey callback for importing a context.
        """
        ImportContextDialog.opener(caller=self)

    @OpenableListControl.Autohotkey("d", "Set as default", True)
    def set_default_context(self, _):
        """
        Hotkey callback for setting the default context.
        """
        if self.selection["type"] == "role":
            Common.Session.set_message(
                "Cannot set a role as a default context, sorry",
                Common.color("message_error"),
            )
            return
        Common.Configuration["default_context"] = self.selection.name
        Common.Configuration.write_config()
        self.reload_contexts()

    @OpenableListControl.Autohotkey("KEY_ENTER", "Select", True)
    def select_context(self, _, initial=False):
        """
        Hotkey callback for setting the active context.
        """
        if self.selection["type"] == "role":
            Common.Session.set_message(
                "Cannot assume roles yet", Common.color("message_error")
            )
            return
        Common.Session.context = self.selection.name
        if Common.Session.context_data["mfa_device"] != "" and (
            "mfa_device" not in Common.Session.context_auth
            or Common.Session.context_auth["mfa_device"]
            != Common.Session.context_data["mfa_device"]
            or "expiry" not in Common.Session.context_auth
            or datetime.datetime.now()
            > datetime.datetime.fromtimestamp(Common.Session.context_auth["expiry"])
        ):
            if initial:
                self.on_become_frame_hooks.append((MFADialog.opener, {"caller": self}))
            else:
                MFADialog.opener(caller=self)

    @OpenableListControl.Autohotkey(ControlCodes.D, "Delete context", True)
    def delete_selected_context(self, _):
        """
        Hotkey callback for deleting the selected context.
        """
        if self.selection.name == "localstack":
            Common.Session.set_message(
                "Cannot delete localstack context", Common.color("message_error")
            )
            return
        if self.selection.name in Common.Configuration.keystore.get_all_ref_targets():
            Common.Session.set_message(
                "Cannot delete context because context is referenced by another context.",
                Common.color("message_error"),
            )
            return
        DeleteResourceDialog.opener(
            caller=self,
            resource_type="context",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
        )

    def do_delete(self, **kwargs):
        """
        Action callback for the context deletion dialog.
        """
        Common.Configuration.delete_context(self.selection["name"])
        self.reload_contexts()



"""
Module for IAM instance profile related resources.
"""

from .base_control import (
    Describer,
    ResourceLister,
    SelectionAttribute,
    SelectorTemplateField,
    TemplateDict,
)
from .termui.ui import ControlCodes


class InstanceProfileDescriber(Describer):
    """
    Describer control for IAM instance Profiles.
    """

    prefix = "instance_profile_browser"
    title = "Instance Profile"

    resource_type = "instance profile"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Instance Profile"
    describe_method = "get_instance_profile"
    describe_kwarg_name = "InstanceProfileName"
    object_path = ".InstanceProfile"

    def __init__(self, *args, entry, entry_key="name", **kwargs):
        from .arn import ARN

        if entry_key == "id":  # id is an Arn from a single relation lister
            arn = ARN(entry["id"])
            entry_key = "name"
            entry["name"] = arn.resource_id_first
        super().__init__(*args, entry=entry, entry_key=entry_key, **kwargs)


def _instance_profile_determine_role_count(instanceprofile):
    return str(len(instanceprofile["Roles"])) if "Roles" in instanceprofile else 0


@ResourceLister.Autocommand("RoleLister", "u", "View instance profile", "role")
class InstanceProfileLister(ResourceLister):
    """
    Lister control for IAM Instance Profiles.
    """

    prefix = "instance_profile_list"
    title = "Instance Profiles"
    command_palette = ["instanceprofile", "instanceprofiles"]

    resource_type = "instance profile"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Instance Profile"
    item_path = ".InstanceProfiles"
    columns = {
        "name": {
            "path": ".InstanceProfileName",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
        "id": {"path": ".InstanceProfileId", "size": 15, "weight": 1},
        "path": {"path": ".Path", "size": 40, "weight": 2},
        "roles": {
            "path": _instance_profile_determine_role_count,
            "size": 5,
            "weight": 3,
        },
        "arn": {"path": ".Arn", "hidden": True},
    }
    describe_command = InstanceProfileDescriber.opener
    open_command = "u"

    def title_info(self):
        return None if self.role is None else f"Role: {self.role['name']}"

    def add_role(self, _):
        """
        Hotkey callback for adding a role to an instance profiel.
        """
        self.selector_template(
            "add_role_to_instance_profile",
            TemplateDict(
                {
                    "InstanceProfileName": SelectionAttribute("name"),
                    "RoleName": SelectorTemplateField(),
                }
            ),
            "Add role to instance profile",
            "add",
            "RoleLister",
            resource_type="group",
            success_template="Adding role {RoleName} to instance profile {InstanceProfileName}",
        )(self.selection)

    def __init__(self, *args, role=None, **kwargs):
        self.role = role
        self.list_method = (
            "list_instance_profiles"
            if self.role is None
            else "list_instance_profiles_for_role"
        )
        self.list_kwargs = {} if self.role is None else {"RoleName": self.role["name"]}
        super().__init__(*args, **kwargs)
        if self.role is None:
            self.add_hotkey("j", self.add_role, "Attach Role", True)
            self.confirm_template(
                "delete_instance_profile",
                TemplateDict(
                    {
                        "InstanceProfileName": SelectionAttribute("name"),
                    }
                ),
                hotkey=ControlCodes.D,
                hotkey_tooltip="Delete Instance Profile",
            )
        else:
            self.confirm_template(
                "remove_role_from_instance_profile",
                TemplateDict(
                    {
                        "RoleName": self.role["name"],
                        "InstanceProfileName": SelectionAttribute("name"),
                    }
                ),
                success_template="Removing role {RoleName} from instance profile {InstanceProfileName}",
                action_name="Detach",
                resource_type="role",
                resource_identifier=self.role["name"],
                from_what="instance profile",
                from_what_name=SelectionAttribute("name"),
                hotkey=ControlCodes.D,
                hotkey_tooltip="Detach Instance Profile",
            )



"""
Module for meta controls.
"""
from .base_control import OpenableListControl
from .common import Common
from .termui.list_control import ListEntry


class CommanderOptionsLister(OpenableListControl):
    """
    List page for available command palette options.

    Attributes
    ----------
    phony : dict
        Deprecated. Legacy map of commands mapped to titles for bare ListControl subclasses.
    """

    title = "Commands"
    prefix = "help"

    def __init__(self, *args, **kwargs):
        """
        Initializes a CommanderOptionsLister.
        """
        super().__init__(*args, **kwargs)
        self.phony = {}
        self.column_titles = {}
        self.column_order = []
        self.add_column("command", 12)
        self.add_column("resource", 48)
        self.reload()

    def reload(self):
        """
        Reloads the list of available commands, repopulating the control.
        """
        self.entries = []
        for cmd in Common.Session.commander_options:
            opt = Common.Session.commander_options[cmd]
            if hasattr(opt, "__self__"):
                list_entry = ListEntry(cmd, command=cmd, resource=opt.__self__.title)
            elif cmd in self.phony:
                list_entry = ListEntry(cmd, command=cmd, resource=self.phony[cmd])
            else:
                list_entry = ListEntry(cmd, command=cmd, resource="")
            list_entry.controller_data["fn"] = opt
            self.entries.append(list_entry)
        self.entries.sort(key=lambda x: x["command"])

    @OpenableListControl.Autohotkey("KEY_ENTER", "Open", True)
    def open(self, _):
        """
        Hotkey callback for opening the list control for the current selection.
        """
        Common.Session.replace_frame(self.selection.controller_data["fn"]())



"""
Module for IAM policy resources.
"""

from .base_control import (
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SelectionAttribute,
    TemplateDict,
)
from .common import Common
from .resource_iam import EMPTY_POLICY_DOCUMENT
from .termui.ui import ControlCodes


class PolicyVersionDescriber(Describer):
    """
    Describer control for a policy version.
    """

    prefix = "policy_version_browser"
    title = "Policy Document"

    resource_type = "policy version"
    main_provider = "iam"
    category = "IAM"
    subcategory = "PolicyVersion"
    describe_method = "get_policy_version"
    object_path = ".PolicyVersion"

    def title_info(self):
        return self.name

    def __init__(self, *args, policy="", **kwargs):
        self.arn = Common.Session.jq(".Arn").input(text=policy).first()
        self.name = Common.Session.jq(".PolicyName").input(text=policy).first()
        version = Common.Session.jq(".DefaultVersionId").input(text=policy).first()
        self.describe_kwargs = {"PolicyArn": self.arn, "VersionId": version}
        kwargs["entry"] = None
        kwargs["entry_key"] = None
        super().__init__(*args, **kwargs)


class PolicyDescriber(Describer):
    """
    Describer control for IAM policy resources.
    """

    prefix = "policy_browser"
    title = "Policy"

    resource_type = "policy"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Managed Policies"
    describe_method = "get_policy"
    describe_kwarg_name = "PolicyArn"
    object_path = ".Policy"
    default_entry_key = "arn"

    def __init__(self, *args, **kwargs):
        self.additional_commands = {
            "d": {
                "command": PolicyVersionDescriber.opener,
                "data_arg": "policy",
                "tooltip": "View Policy Document",
                "kwargs": {"entry": None},
            }
        }
        super().__init__(*args, **kwargs)


@ResourceLister.Autocommand("UserLister", "m", "View policies", "user")
@ResourceLister.Autocommand("GroupLister", "m", "View policies", "group")
@ResourceLister.Autocommand("RoleLister", "m", "View policies", "role")
class PolicyLister(ResourceLister):
    """
    Lister control for IAM policy resources.
    """

    prefix = "policy_list"
    title = "Policies"
    command_palette = ["policy", "policies"]

    resource_type = "policy"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Managed Policies"
    columns = {
        "name": {
            "path": ".PolicyName",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
        "arn": {"path": ".PolicyArn", "size": 80, "weight": 1},
    }
    describe_command = PolicyDescriber.opener

    def title_info(self):
        if self.user is not None:
            return f"User: {self.user['name']}"
        if self.group is not None:
            return f"Group: {self.group['name']}"
        if self.role is not None:
            return f"Role: {self.role['name']}"
        return None

    def get_scope_tooltip(self):
        """
        Hotkey tooltip callback for policy scopes.
        """
        return f"Change Scope ({self.scope})"

    def change_scope(self, *args):
        """
        Hotkey callback for changing the viewed policy scopes.
        """
        idx = self.scopes.index(self.scope)
        newidx = idx + 1
        if newidx >= len(self.scopes):
            newidx = 0
        self.scope = self.scopes[newidx]
        self.list_kwargs["Scope"] = self.scope
        Common.Session.set_message("Refreshing...", Common.color("message_info"))
        self.refresh_data()

    def create_policy(self, *args):
        """
        Hotkey callback for creating a new policy.
        """
        creator = ListResourceDocumentCreator(
            "iam",
            "create_policy",
            None,
            as_json=["Tags"],
            initial_document={
                "PolicyName": "",
                "Path": "/",
                "Description": "",
                "Tags": [{"Key": "", "Value": ""}],
                **EMPTY_POLICY_DOCUMENT,
            },
        )
        creator.edit()
        self.refresh_data()

    def delete_policy(self, _):
        """
        Hotkey callback for deleting an existing policy. Additional validation is required for this one.
        """
        if "attached" not in self.selection or int(self.selection["attached"]) > 0:
            Common.Session.set_message(
                "Policy must be detached before deletion", Common.color("message_error")
            )
            return
        self.confirm_template(
            "delete_policy", TemplateDict({"PolicyArn": SelectionAttribute("arn")})
        )(self.selection)

    def detach_policy(self, _):
        """
        Hotkey callback for detaching the policy from the contextual IAM datatype.
        """
        if self.user is not None:
            dtype = "user"
            dname = self.user.name
        elif self.group is not None:
            dtype = "group"
            dname = self.group.name
        elif self.role is not None:
            dtype = "role"
            dname = self.role.name
        else:
            return
        argname = f"{dtype.capitalize()}Name"
        self.confirm_template(
            f"detach_{dtype}_policy",
            TemplateDict(
                {"PolicyArn": SelectionAttribute("arn"), argname: dname},
            ),
            action_name="Detach",
            from_what=dtype,
            from_what_name=dname,
        )(self.selection)

    def __init__(self, *args, user=None, group=None, role=None, **kwargs):
        isset = int(user is not None) + int(group is not None) + int(role is not None)
        if isset > 1:
            raise RuntimeError("Only one of user, group or role can be set.")
        self.user = user
        self.group = group
        self.role = role
        self.resource_key = "iam"
        self.scopes = ["All", "AWS", "Local"]
        self.scope = "All"
        if "selector_cb" in kwargs:
            self.primary_key = "arn"

        if user is not None:
            self.list_method = "list_attached_user_policies"
            self.list_kwargs = {"UserName": self.user["name"]}
            self.item_path = ".AttachedPolicies"
        elif group is not None:
            self.list_method = "list_attached_group_policies"
            self.list_kwargs = {"GroupName": self.group["name"]}
            self.item_path = ".AttachedPolicies"
        elif role is not None:
            self.list_method = "list_attached_role_policies"
            self.list_kwargs = {"RoleName": self.role["name"]}
            self.item_path = ".AttachedPolicies"
        else:
            self.list_method = "list_policies"
            self.list_kwargs = {"Scope": self.scope}
            self.item_path = ".Policies"
            self.columns["arn"]["path"] = ".Arn"
            self.columns["attached"] = {
                "path": ".AttachmentCount",
                "weight": 2,
                "size": 8,
            }

        super().__init__(*args, **kwargs)
        if isset == 0:
            self.add_hotkey("n", self.create_policy, "Create new policy")
            self.add_hotkey("s", self.change_scope, self.get_scope_tooltip)
            self.add_hotkey(ControlCodes.D, self.delete_policy, "Delete policy", True)
        else:
            self.add_hotkey(ControlCodes.D, self.detach_policy, "Detach policy", True)



"""
This module defines list controls for displaying data in tables.
"""
import datetime

from .color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .common import Commons
from .control import HotkeyControl


class ListControl(HotkeyControl):
    """
    A ListControl is a large control which allows displaying any amount of entries as a list of rows, where each row is a set of columns.
    ListControls allow navigation and selection of items which in turn allows the use of hotkeys for operating on entries.

    Attributes
    ----------
    entries : list(awsc.termui.list_control.ListEntry)
        A list of entries.
    color : awsc.termui.color.Color
        The color of unselected list entries.
    selection_color : awsc.termui.color.Color
        The color of selected list entries.
    title_color : awsc.termui.color.Color
        The color of the title in the border bar.
    selected : int
        The index of the selected row.
    column_titles : dict
        A mapping of column titles to the size of the column. By default, this contains the initial size of the column. Space will be given to
        and taken away from columns roughly equally based on window size.
    column_order : list
        The order in which columns should be displayed. A column must be present in both column_order and column_titles to function properly.
    calculated : int
        Cached value for the block width for which column sizes have been calculated.
    _filter : str
        The currently set filter for filtering list entries.
    _cache : list
        Cache for the list of filtered entries.
    top : int
        How many entries should be skipped before displaying (how far we're scrolled down).
    update_color : awsc.termui.color.Color
        The color of rows that have been recently (in the past 5 seconds) updated.
    update_selection_color : awsc.termui.color.Color
        The color of rows that have been recently (in the past 5 seconds) updated and are selected.
    """

    def __init__(
        self,
        *args,
        color=ColorGold,
        selection_color=ColorBlackOnGold,
        title_color=ColorBlackOnOrange,
        update_color=ColorGold,
        update_selection_color=ColorBlackOnGold,
        **kwargs,
    ):
        """
        Initializes a ListControl object.
        """
        super().__init__(*args, **kwargs)
        self.entries = []
        self.color = color
        self.selection_color = selection_color
        self.title_color = title_color
        self.selected = 0
        self.column_titles = {"name": 30}
        self.column_order = ["name"]
        self.calculated = 0
        self._filter = None
        self.thread_share["clear"] = False
        self._cache = None
        self.thread_share["new_entries"] = []
        self.thread_share["finalize"] = False
        self.top = 0
        self.update_color = update_color
        self.update_selection_color = update_selection_color

    def validate_hotkey(self, key):
        return self.selection is not None

    @property
    def filter(self):
        """
        Allows for filtering the list entries without removing them. Only entries matching the filter are displayed.

        The filter will do a search in all fields of the entry, even hidden ones.
        """
        return self._filter

    @filter.setter
    def filter(self, value):
        Commons.UIInstance.dirty = True
        self._filter = value
        self.top = 0
        self.selected = 0
        self._cache = None

    def add_entry(self, entry):
        """
        Adds a new entry to the ListControl.

        Parameters
        ----------
        entry : awsc.termui.list_control.ListEntry
            The new entry to add.
        """
        self.entries.append(entry)
        self.sort()
        self._cache = None
        Commons.UIInstance.dirty = True

    @property
    def filtered(self):
        """
        Read-only property that returns the list of filtered entris.
        """
        if self._cache is None:
            self._cache = [e for e in self.entries if e.matches_filter(self._filter)]
        return self._cache

    @HotkeyControl.Autohotkey("KEY_DOWN")
    def list_down(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_DOWN. Moves the selection cursor down by one row.
        """
        if self.selected < len(self.filtered) - 1:
            self.selected += 1
        rows = self.rows
        if self.selected >= self.top + rows:
            self.top = max(0, self.selected - rows + 1)

    @HotkeyControl.Autohotkey("KEY_UP")
    def list_up(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_UP. Moves the selection cursor down by one row.
        """
        if self.selected > 0:
            self.selected -= 1
        if self.selected < self.top:
            self.top = self.selected

    @HotkeyControl.Autohotkey("KEY_PGDOWN")
    def pagedown(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_PAGEDOWN. Moves the selection cursor down by one page - which is the number of visible rows.
        """
        self.selected += self.rows
        self.selected = min(len(self.filtered) - 1, self.selected)
        rows = self.rows
        if self.selected >= self.top + rows:
            self.top = max(0, self.selected - rows + 1)

    @HotkeyControl.Autohotkey("KEY_PGUP")
    def pageup(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_PAGEUP. Moves the selection cursor up by one page - which is the number of visible rows.
        """
        self.selected -= self.rows
        self.selected = max(0, self.selected)
        if self.selected < self.top:
            self.top = self.selected

    @HotkeyControl.Autohotkey("KEY_HOME")
    def home(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_HOME. Moves the selection cursor up to the first entry.
        """
        self.selected = 0
        self.top = 0

    @HotkeyControl.Autohotkey("KEY_END")
    def end(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Hotkey callback for KEY_END. Moves the selection cursor down to the last entry.
        """
        self.selected = len(self.filtered) - 1
        self.top = max(0, self.selected - self.rows + 1)

    @property
    def selection(self):
        """
        Read-only property which contains the currently selected list entry.

        Returns
        -------
        awsc.termui.list_control.ListEntry
            The currently selected ListEntry.
        """
        if len(self.filtered) > self.selected:
            return self.filtered[self.selected]
        return None

    @property
    def rows(self):
        """
        Read-only property which contains the number of visible rows in the control.

        Returns
        -------
        int
            Number of list entries that can be displayed at once.
        """
        corners = self.corners
        y0 = corners[1][0] + (0 if self.border is None else 1)
        y1 = corners[1][1] - (0 if self.border is None else 1)
        return y1 - y0

    def add_column(self, column, min_size=0, index=None):
        """
        Adds a new column to the ListControl.

        Parameters
        ----------
        column : str
            Name of the column to add.
        min_size : int, default=0
            The initial size of the column.
        index : int, optional
            The position where the column should be inserted. By default, the column is added as the last column.
        """
        if index is None:
            self.column_order.append(column)
        else:
            self.column_order.insert(index, column)
        self.column_titles[column] = min_size
        self.calculated = 0
        Commons.UIInstance.dirty = True

    def before_paint_critical(self):
        """
        Critical section for the pre-paint hook. This function holds the control mutex for its entire runtime.

        This function will call other critical section functions for handlers.
        """
        if self.thread_share["clear"]:
            Commons.UIInstance.dirty = True
            self.entries = []
            self.top = 0
            self.thread_share["clear"] = False
        if len(self.thread_share["new_entries"]) > 0:
            Commons.UIInstance.dirty = True
            self.handle_new_entries_critical(self.thread_share["new_entries"])
            self._cache = None
            self.thread_share["new_entries"] = []
        if self.thread_share["finalize"]:
            self.handle_finalization_critical()
            self.thread_share["finalize"] = False

    def handle_new_entries_critical(self, entries):
        """
        Critical section for the new entry hook. Called by before_paint_critical, therefore holds the control mutex.

        Parameters
        ----------
        entries : list(awsc.termui.list_control.ListEntry)
            A list of new entries to add to the control.
        """
        self.entries.extend(entries)
        self.sort()

    def handle_finalization_critical(self):
        """
        Critical section for the finalization of new entries. Called by before_paint_critical, therefore holds the control mutex.
        """

    def before_paint(self):
        super().before_paint()
        with self.mutex:
            self.before_paint_critical()
        # if self.selected >= len(self.entries):
        #  self.selected = len(self.entries) - 1
        # if self.selected < 0:
        #  self.selected = 0

    def sort(self):
        """
        Sorts the list of entries. Expected to be overridden by controls which actually understand how they want to sort their entries.

        Sorting happens automatically each time new entries are inserted.
        """

    def paint(self):
        win = self.w_in
        if win != self.calculated:
            vals = 0
            for _, value in self.column_titles.items():
                vals += value
            if vals > win:
                ratio = float(vals) / float(win)
                tot = 0
                # pylint: disable=consider-iterating-dictionary # Operation modifies dictionary
                for k in self.column_titles.keys():
                    self.column_titles[k] = int(float(self.column_titles[k]) / ratio)
                    tot += self.column_titles[k]
                rem = win - tot
                if "name" in self.column_titles:
                    self.column_titles["name"] += rem
                else:
                    self.column_titles[self.column_order[0]] += rem
            elif vals < win:
                diff = win - vals
                part = int(diff / len(self.column_titles))
                rem = diff - (part * len(self.column_titles))
                # pylint: disable=consider-iterating-dictionary # Operation modifies dictionary
                for k in self.column_titles.keys():
                    self.column_titles[k] += part
                if "name" in self.column_titles:
                    self.column_titles["name"] += rem
                else:
                    self.column_titles[self.column_order[0]] += rem
            self.calculated = win

        super().paint()
        corners = self.corners
        y = corners[1][0] + (0 if self.border is None else 1)
        x = corners[0][0] + (0 if self.border is None else 1)
        now = datetime.datetime.now()
        rows = self.rows
        for col in self.column_order:
            text = col
            maximum = self.column_titles[col]
            if len(col) > maximum:
                text = col[:maximum]
            elif len(col) < maximum:
                text = col + (int(maximum - len(col)) * " ")
            Commons.UIInstance.print(text.upper(), xy=(x, y), color=self.title_color)
            x += maximum
        y += 1
        for i in range(self.top, min(len(self.filtered), self.top + rows)):
            item = self.filtered[i]
            x = corners[0][0] + (0 if self.border is None else 1)
            for col in self.column_order:
                text = item[col]
                if text is None:
                    import sys

                    print(f"None encountered! Row: {item}", file=sys.stderr)
                maximum = self.column_titles[col]
                if len(text) > maximum:
                    text = text[:maximum]
                elif len(text) < maximum:
                    text = text + ((maximum - len(text)) * " ")
                if i == self.selected:
                    if now - item.updated < datetime.timedelta(seconds=3):
                        color = self.update_selection_color
                    else:
                        color = self.selection_color
                else:
                    if now - item.updated < datetime.timedelta(seconds=3):
                        color = self.update_color
                    else:
                        color = self.color
                Commons.UIInstance.print(text, xy=(x, y), color=color)
                x += maximum
            y += 1


class ListEntry(dict):
    """
    Represents a single entry in a ListControl. A fancy dict, if you will.

    Attributes
    ----------
    columns : dict
        A mapping of column name to column value.
    controller_data : object
        Arbitrary object which allows controlling ListControl derivative objects to store metadata associated with the list entry.
    updated : datetime.datetime
        The time this entry was last updated, to track when entries should be displayed as updated.
    """

    def __init__(self, name, controller_data=None, **kwargs):
        """
        Initializes a ListEntry.

        Parameters
        ----------
        name: str
            The initial value for the name field.
        controller_data : object, optional
            Arbitrary data.
        **kwargs : dict
            Each additional keyword argument will be treated as values for fields.
        """
        super().__init__(
            **{k: str(v) if v is not None else "" for k, v in kwargs.items()}
        )
        self.name = name
        if controller_data is None:
            self.controller_data = {}
        else:
            self.controller_data = controller_data
        self.updated = datetime.datetime.now()

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key in ("controller_data", "updated"):
            super().__setattr__(key, value)
        else:
            self[key] = value

    def __setitem__(self, key, value):
        if value is None:
            value = ""
        super().__setitem__(key, value)

    def matches_filter(self, filt):
        """
        Decider for whether the list entry matches a particular filter.

        Parameters
        ----------
        filt : str
            The filter to check.

        Returns
        -------
        bool
            Whether the list entry matches the given filter.
        """
        if filt is None:
            return True
        filt = filt.lower()
        for _, column in self.items():
            if filt in column.lower():
                return True
        return False

    def mutate(self, other):
        """
        Transforms a list entry into another list entry.
        Used by ListControl derivatives to update entries in-place based on periodic API queries, maintaining the updated timestamp.

        Parameters
        ----------
        other : awsc.termui.list_control.ListEntry
            The target ListEntry to transform this ListEntry into.
        """
        updated = False
        if self.name != other.name or len(self) != len(other):
            updated = True
        else:
            for col in self:
                if col not in other or self[col] != other[col]:
                    updated = True
                    break
        if not updated:
            return
        self.clear()
        self.update(other)
        self.updated = datetime.datetime.now()
        self.controller_data = other.controller_data



"""
Module for IAM user resources.
"""

import pyperclip

from .base_control import (
    DeleteResourceDialog,
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SelectionAttribute,
    SelectorTemplateField,
    SingleNameDialog,
    TemplateDict,
)
from .common import Common
from .resource_iam import InlinePolicyAttacher
from .termui.ui import ControlCodes


class UserDescriber(Describer):
    """
    Describer control for user resources.
    """

    prefix = "user_browser"
    title = "User"

    resource_type = "user"
    main_provider = "iam"
    category = "IAM"
    subcategory = "User"
    describe_method = "get_user"
    describe_kwarg_name = "UserName"
    object_path = ".User"


@ResourceLister.Autocommand("GroupLister", "u", "List group users", "group")
class UserLister(ResourceLister):
    """
    Lister control for user resources.
    """

    prefix = "user_list"
    title = "Users"
    command_palette = ["user", "users"]

    resource_type = "user"
    main_provider = "iam"
    category = "IAM"
    subcategory = "User"
    item_path = ".Users"
    columns = {
        "name": {
            "path": ".UserName",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
        "path": {"path": ".Path", "size": 40, "weight": 1},
        "last login": {"path": ".PasswordLastUsed", "size": 15, "weight": 2},
        "arn": {"path": ".Arn", "hidden": True},
    }
    describe_command = UserDescriber.opener
    open_command = "m"

    def title_info(self):
        return f"Group: {self.group['name']}" if self.group is not None else None

    def create_user(self, _):
        """
        Hotkey callback for creating a new user.
        """
        creator = ListResourceDocumentCreator(
            "iam",
            "create_user",
            None,
            as_json=["Tags"],
            initial_document={
                "UserName": "",
                "Path": "/",
                "PermissionsBoundary": "",
                "Tags": [{"Key": "", "Value": ""}],
            },
        )
        creator.edit()
        self.refresh_data()

    def delete_user(self, _):
        """
        Hotkey callback for deleting a user.
        """
        self.confirm_template(
            "delete_user",
            TemplateDict(
                {
                    "UserName": SelectionAttribute("name"),
                }
            ),
        )(self.selection)

    @ResourceLister.Autohotkey("a", "Attach Policy", True)
    def attach_policy(self, _):
        """
        Hotkey callback for attaching a policy to a user.
        """
        self.selector_template(
            "attach_user_policy",
            TemplateDict(
                {
                    "UserName": SelectionAttribute("name"),
                    "PolicyArn": SelectorTemplateField(),
                }
            ),
            "Attach policy to user",
            "attach",
            "PolicyLister",
            resource_type="policy",
            success_template="Attaching policy {PolicyArn} to user {UserName}",
        )(self.selection)

    def add_user_to_group(self, _):
        """
        Hotkey callback for adding a user to a group.
        """
        self.selector_template(
            "add_user_to_group",
            TemplateDict(
                {
                    "UserName": SelectionAttribute("name"),
                    "GroupName": SelectorTemplateField(),
                }
            ),
            "Add user to group",
            "add",
            "GroupLister",
            resource_type="group",
            success_template="Adding user {UserName} to group {GroupName}",
        )(self.selection)

    def remove_from_group(self, _):
        """
        Hotkey callback for removing a user from a group.
        """
        self.confirm_template(
            "remove_user_from_group",
            TemplateDict(
                {
                    "UserName": SelectionAttribute("name"),
                    "GroupName": self.group["name"],
                }
            ),
            from_what="group",
            from_what_name=self.group["name"],
            action_name="Remove",
        )(self.selection)

    def determine_login_profile(self, user):
        """
        Check if a user has a login profile.

        Parameters
        ----------
        user : str
            The username to check.

        Returns
        -------
        bool
            True if the user has a login profile.
        """
        resp = Common.generic_api_call(
            "iam",
            "get_login_profile",
            {"UserName": user},
            "Get Login Profile",
            "IAM",
            subcategory="User",
            resource=user,
        )
        if resp["Success"]:
            return True
        return False

    @ResourceLister.Autohotkey("p", "Get/create login profile", True)
    def get_or_create_login_profile(self, _):
        """
        Hotkey callback for retrieving the user login profile. If one doesn't exist, a dialog is presented to create one.
        """
        has_login_profile = self.determine_login_profile(self.selection["name"])
        if has_login_profile:
            Common.Session.push_frame(
                LoginProfileDescriber.opener(
                    entry=self.selection, caller=self, pushed=True
                )
            )
        else:
            SingleNameDialog(
                self.parent,
                f"Create login profile for {self.selection['name']}",
                self.do_create_login_profile,
                label="Password:",
                what="password",
                subject="login profile",
                default="",
                caller=self,
            )

    def do_create_login_profile(self, password):
        """
        Action callback for the login profile creation dialog.

        Parameters
        ----------
        password : str
            Password for the new login profile.
        """
        resp = Common.generic_api_call(
            "iam",
            "create_login_profile",
            {
                "UserName": self.selection["name"],
                "Password": password,
                "PasswordResetRequired": True,
            },
            "Create Login Profile",
            "IAM",
            subcategory="User",
            resource=self.selection["name"],
        )
        if resp["Success"]:
            Common.Session.push_frame(
                LoginProfileDescriber.opener(
                    entry=self.selection, caller=self, pushed=True
                )
            )

    def __init__(self, *args, group=None, **kwargs):
        self.group = group
        self.list_method = "list_users" if group is None else "get_group"
        self.list_kwargs = {"GroupName": group["name"]} if group is not None else {}
        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "user")
        self.add_hotkey("a", self.attach_policy, "Attach Policy", True)
        if self.group is None:
            self.add_hotkey("j", self.add_user_to_group, "Add to Group", True)
            self.add_hotkey(ControlCodes.D, self.delete_user, "Delete User", True)
            self.add_hotkey("n", self.create_user, "Create User")
        else:
            self.add_hotkey(
                ControlCodes.D, self.remove_from_group, "Remove from Group", True
            )


class LoginProfileDescriber(Describer):
    """
    Describer control for login profiles.
    """

    prefix = "login_profile_browser"
    title = "Login Profile"

    def title_info(self):
        return f"User: {self.user['name']}"

    def __init__(self, *args, entry, **kwargs):
        self.user = entry
        self.resource_key = "iam"
        self.describe_method = "get_login_profile"
        self.describe_kwarg_name = "UserName"
        self.object_path = ".LoginProfile"
        super().__init__(*args, entry=entry, **kwargs)
        self.add_hotkey(
            ControlCodes.D, self.delete_login_profile, "Delete Login Profile"
        )

    # TODO: Create implementation of confirm_template for Describers.
    @Describer.Autohotkey(ControlCodes.D, "Delete login profile")
    def delete_login_profile(self, _):
        """
        Hotkey callback for deleting a login profile.
        """
        DeleteResourceDialog.opener(
            caller=self,
            resource_type="login profile",
            resource_identifier=self.user["name"],
            from_what="user",
            from_what_name=self.user["name"],
            callback=self.do_delete_login_profile,
        )

    def do_delete_login_profile(self, *args, **kwargs):
        """
        Action callback for deleting a login profile.
        """
        resp = Common.generic_api_call(
            "iam",
            "delete_login_profile",
            {
                "UserName": self.user["name"],
            },
            "Delete Login Profile",
            "IAM",
            subcategory="User",
            resource=self.user["name"],
            success_template="Deleting login profile for {UserName}",
        )
        if resp["Success"]:
            Common.Session.pop_frame()


class MFADeviceLister(ResourceLister):
    prefix = "mfa_list"
    title = "MFA Devices"
    command_palette = ["mfa"]

    resource_type = "mfa device"
    main_provider = "iam"
    category = "IAM"
    subcategory = "MFA Device"
    list_method = "list_mfa_devices"
    item_path = ".MFADevices"
    columns = {
        "username": {
            "path": ".UserName",
            "size": 25,
            "weight": 0,
        },
        "serial number": {
            "path": ".SerialNumber",
            "size": 100,
            "weight": 1,
            "sort_weight": 0,
        },
        "enable date": {
            "path": ".EnableDate",
            "size": 20,
            "weight": 2,
        },
        "name": {
            "path": ".SerialNumber",
            "hidden": True,
        },
    }

    @ResourceLister.Autohotkey("r", "Copy Serial Number", True)
    def copy_sn(self, *args):
        """
        Copies the serial number of the selection to the clipboard if available.
        """
        pyperclip.copy(self.selection["serial number"])
        Common.Session.set_message(
            "Copied resource serial number to clipboard",
            Common.color("message_success"),
        )

    # @ResourceLister.Autohotkey(ControlCodes.D, 'Deactivate MFA device', True)
    # def
    @ResourceLister.Autohotkey(ControlCodes.A, "Associate", True)
    def associate(self, *args):
        """
        Associates the MFA devices with the currently enabled context.
        """
        Common.Configuration["contexts"][Common.Session.context][
            "mfa_device"
        ] = self.selection["serial number"]
        Common.Configuration.write_config()



"""
Module for logging-related controls.
"""
import datetime
import json
from operator import attrgetter

from .base_control import GenericDescriber, OpenableListControl, datetime_hack
from .common import Common
from .termui.list_control import ListEntry


class LogViewer(GenericDescriber):
    """
    Browser control for viewing log entries.
    """

    def __init__(self, *args, selection, **kwargs):
        columns = selection.copy()
        columns["context"] = json.loads(columns["context"])
        content = json.dumps(columns, default=datetime_hack, indent=2, sort_keys=True)
        super().__init__(
            *args,
            describing="logs",
            content=content,
            **kwargs,
        )


class LogLister(OpenableListControl):
    """
    Lister control for log entries.
    """

    prefix = "log"
    title = "Logs"
    describer = LogViewer.opener

    def __init__(self, parent, alignment, dimensions, *args, **kwargs):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("ssh_key_list_generic", "generic"),
            selection_color=Common.color("ssh_key_list_selection", "selection"),
            title_color=Common.color("ssh_key_list_heading", "column_title"),
            *args,
            **kwargs,
        )

        self.add_column("type", 12)
        self.add_column("category", 20)
        self.add_column("subcategory", 20)
        self.add_column("resource", 20)
        self.add_column("timestamp", 20)
        self.logholder = Common._logholder
        self.logholder.attach(self)

    def add_raw_entry(self, entry):
        """
        Inserts a new raw entry. Callback for the logholder for pushing entries into the lister.
        """
        self.add_entry(
            ListEntry(
                entry["summary"],
                **{
                    "category": entry["category"],
                    "subcategory": entry["subcategory"]
                    if "subcategory" in entry
                    and entry["subcategory"] is not None
                    and entry["subcategory"] != "null"
                    else "<n/a>",
                    "resource": entry["resource"]
                    if entry["resource"] is not None
                    else "",
                    "type": entry["type"],
                    "timestamp": datetime.datetime.utcfromtimestamp(
                        entry["timestamp"]
                    ).strftime("%Y-%m-%d %H:%M:%S"),
                    "raw_timestamp": entry["timestamp"],
                    "message": entry["message"],
                    "context": entry["context"] if "context" in entry else {},
                },
            )
        )

    def sort(self):
        self.entries.sort(reverse=True, key=attrgetter("raw_timestamp"))
        self._cache = None

    def on_close(self):
        """
        Cleanup hook. Detaches the control from the log storage.
        """
        self.logholder.detach()



"""
Session manager module.
"""
import json
import os
import subprocess
import tempfile
import time
import urllib.error
import urllib.request

import jq
from packaging import version

from .info import InfoDisplay, NeutralDialog
from .termui.alignment import BottomLeftAnchor, Dimension, TopLeftAnchor
from .termui.dialog import DialogFieldLabel
from .termui.ui import UI, ControlCodes
from .version import VERSION as current_version


class Session:
    """
    Session manager object. Manages what's visible on the screen. Also stores some information related to the current user session.

    Attributes
    ----------
    jqc : dict
        JQ cache. Stores compiled JQ objects keyed by JQ statement.
    ui : awsc.termui.ui.UI
        The main UI object of AWSC.
    config : awsc.config.config.Config
        The configuration object also stored in Common. We cannot import Common.
    context_update_hooks : list(callable)
        A list of hooks to run when the active context is updated.
    info_display : awsc.info.InfoDisplay
        The top left information display, which contains information about the active session.
    global_hotkey_tooltips : dict
        A dict of global hotkeys and associated tooltips.
    message_display : awsc.info.NeutralDialog
        The message display is a dialog on the backend which shows a flashed message to the user.
    _message_labels : list(awsc.termui.dialog.DialogFieldLabel)
        Label fields for each line of the message display.
    _context : str
        Current AWS context.
    _region : str
        Current AWS region.
    _ssh_key : str
        Current SSH key.
    stack : list
        A stack of stack frames.
    stack_frame : list
        The topmost stack frame. This frame is the one that is being displayed to the user.
    resource_main : awsc.termui.control.Control
        The main control of the stack frame. Usually the first element passed in a new frame.
    service_provider : awsc.aws.AWS
        The AWS connection provider.
    message_time : float
        The remaining amount of time for which a flashed message continues to be displayed.
    last_tick : time.time
        The timestamp when the last tick was executed, for deltatime purposes.
    filterer: awsc.commander.Filterer
        The filterer object that is open in the session, if there is one.
    commander: awsc.commander.Commander
        The commander object that is open in the session, if there is one.
    commander_options : dict(str, callable)
        A map of available commander options mapped to the callback for that command.
    control_registry : dict(str, callable)
        A map of available awsc control opener functions keyed by classname.
    common : awsc.common.Common
        Reference back to the common object.
    """

    def jq(self, stmt):
        """
        Creates a compiled JQ object, or fetches from cache if it exists.

        Parameters
        ----------
        stmt : str
            The JQ statement to compile.

        Returns
        -------
        jq._Program
            The compiled JQ statement.
        """
        if stmt not in self.jqc:
            self.jqc[stmt] = jq.compile(stmt)
        return self.jqc[stmt]

    def __init__(self, config, highlight_color, generic_color, common):
        self.jqc = {}
        self.ui = UI()
        self.config = config
        self.context_update_hooks = []
        self.info_display = InfoDisplay(
            self.ui.top_block,
            TopLeftAnchor(1, 0),
            Dimension("50%", 8),
            highlight_color=highlight_color,
            generic_color=generic_color,
            info=[
                "AWSC Version",
                "Context",
                "Region",
                "UserId",
                "Account",
                "MFA",
                "SSH Key",
                "Default SSH username",
            ],
            weight=-10,
        )
        self.global_hotkey_tooltips = {
            ":": "Command palette",
            "/": "Filter / search",
            ControlCodes.C: "Exit",
        }
        self.message_display = NeutralDialog(
            self.ui.top_block,
            BottomLeftAnchor(0, 0),
            Dimension("100%", 3),
            confirm_text=None,
            cancel_text=None,
            color=generic_color,
            weight=-10,
        )
        self._message_labels = [DialogFieldLabel(""), DialogFieldLabel("")]
        for elem in self._message_labels:
            self.message_display.add_field(elem)
        self._context = None
        self.context = config["default_context"]
        self.region = config["default_region"]
        self.ssh_key = config["default_ssh_key"]
        self.stack = []
        self.stack_frame = []
        self.resource_main = None
        self.service_provider = None
        self.message_time = 0
        self.last_tick = time.time()
        self.ui.tickers.append(self.tick)

        self.filterer = None
        self.commander = None
        self.commander_options = {}
        self.control_registry = {}
        self.common = common

        self.set_version_information()

    def set_version_information(self):
        """
        Updates the version information on the info display.
        """
        update_character = "↑"
        version_str = current_version
        latest_tag = version.parse(current_version)
        latest_stable = latest_tag
        current_tag = latest_tag
        try:
            with urllib.request.urlopen("https://pypi.org/pypi/awsc/json") as resp:
                if resp.status == 200:
                    package_info = resp.read()
                    package_info_json = json.loads(package_info)
                    for release in package_info_json["releases"].keys():
                        release_tag = version.parse(release)
                        if release_tag > latest_tag:
                            latest_tag = release_tag
                        if (
                            not release_tag.is_prerelease
                            and release_tag > latest_stable
                        ):
                            latest_stable = release_tag
                    if latest_tag != current_tag:
                        if latest_stable != current_tag:
                            version_str = f"{str(current_tag)} ({update_character}{str(latest_stable)}, {update_character}?{str(latest_tag)}"
                        else:
                            version_str = f"{str(current_tag)} ({update_character}?{str(latest_tag)}"
                    else:
                        version_str = str(current_tag)

            self.info_display["AWSC Version"] = version_str
        except urllib.error.URLError as error:
            self.common.log_exception(
                error, "Core", subcategory="Version Check", set_message=False
            )

    def set_message(self, text, color):
        """
        Flashes a message to the user.

        Parameters
        ----------
        text : str
            The text to show.
        color : awsc.termui.color.Color
            The color to use for the shown text.
        """
        max_lines = len(self._message_labels)
        lines = [""] * max_lines
        line_no = 0
        while len(text) > self.ui.width - 1:
            line = text[: self.ui.width - 1]
            lines[line_no] = line
            text = text[self.ui.width - 1 :]
            line_no += 1
            if line_no >= max_lines:
                break
        if line_no < max_lines:
            lines[line_no] = text
        for i in range(max_lines):
            self._message_labels[i].texts = [(lines[i], color)]
        self.message_time = min(len(text) / 10.0, 5.0)
        self.last_tick = time.time()
        self.ui.dirty = True

    def replace_frame(self, new_frame, drop_stack=True):
        """
        Removes the topmost stack frame, replacing it with a new frame.

        Parameters
        ----------
        new_frame : list
            The new frame to replace the topmost stack frame with.
        drop_stack : bool
            If set, drop the entire stack of frames while doing it.
        """
        for elem in self.stack_frame:
            elem.parent.remove_block(elem)
        if drop_stack:
            self.stack = []
        self.resource_main = new_frame[0]
        self.stack_frame = new_frame[:]
        self.ui.dirty = True
        for elem in self.stack_frame:
            if hasattr(elem, "on_become_frame"):
                elem.on_become_frame()

    def push_frame(self, new_frame):
        """
        Pushes a new frame to the stack, replacing the currently active frame for display.

        Calling push_frame, then pop_frame should return to the same state.

        Parameters
        ----------
        new_frame : list
            The new stack frame to add.
        """
        self.stack.append(self.stack_frame[:])
        self.replace_frame(new_frame, drop_stack=False)
        if hasattr(new_frame[0], "add_hotkey"):
            if "KEY_ESCAPE" not in new_frame[0].tooltips:
                new_frame[0].add_hotkey("KEY_ESCAPE", self.pop_frame, "Back")
        self.ui.dirty = True
        for elem in self.stack_frame:
            if hasattr(elem, "on_become_frame"):
                elem.on_become_frame()

    def pop_frame(
        self, *args
    ):  # pylint: disable=unused-argument # hotkey hooks will always be passed an extra argument
        """
        Pops the topmost frame of the stack. Calls on_close of each element of the popped stack frame. Replaces the stack frame with the
        next frame in the stack, if such a frame exists.
        """
        for elem in self.stack_frame:
            if hasattr(elem, "on_close"):
                elem.on_close()
        if len(self.stack) > 0:
            self.replace_frame(self.stack.pop(), drop_stack=False)
        for elem in self.stack_frame:
            elem.reparent()
        self.ui.dirty = True

    def extend_frame(self, control):
        """
        Adds an additional control to the current frame.

        Parameters
        ----------
        control : awsc.termui.control.Control
            The control to insert into the frame.
        """
        self.stack_frame.append(control)
        self.ui.dirty = True
        if hasattr(control, "on_become_frame"):
            control.on_become_frame()

    def remove_from_frame(self, control):
        """
        Removes a control from the current frame. Calls on_close of the removed control.

        Parameters
        ----------
        control : awsc.termui.control.Control
            The control to remove from the frame.
        """
        if hasattr(control, "on_close"):
            control.on_close()
        try:
            self.stack_frame.remove(control)
        except ValueError:
            for elem in reversed(self.stack):
                if control in elem:
                    elem.remove(control)
                    return
        self.ui.dirty = True

    def tick(self):
        """
        Ticker function. Called on every main loop cycle.
        """
        delta = time.time() - self.last_tick
        self.last_tick += delta
        if self.message_time > 0:
            self.message_time -= delta
            if self.message_time <= 0:
                for label in self._message_labels:
                    label.texts = []
        if hasattr(self.resource_main, "auto_refresh"):
            self.resource_main.auto_refresh()

    @property
    def context(self):
        """
        Property. The currently active context.
        """
        return self._context

    @context.setter
    def context(self, value):
        self._context = value
        self.info_display["Context"] = value
        data = self.context_data
        auth = self.context_auth
        self.info_display["MFA"] = (
            "Disabled"
            if data["mfa_device"] == ""
            else ("Authenticated" if "session" in auth else "Unauthenticated")
        )
        for elem in self.context_update_hooks:
            elem()

    @property
    def context_data(self):
        """
        Property. Directly accesses context data for convenience.
        """
        if self._context != "":
            return self.config["contexts"][self._context]
        return {}

    @property
    def context_auth(self):
        """
        Property. Directly accesses context keystore for convenience.
        """
        if self._context != "":
            return self.config.keystore[self._context]
        return {}

    @property
    def context_perm_auth(self):
        """
        Property. Directly accesses context keystore for convenience.
        """
        if self._context != "":
            return self.config.keystore.get_permanent_credentials(self._context)
        return {}

    @property
    def region(self):
        """
        Property. The current region.
        """
        return self._region

    @region.setter
    def region(self, value):
        self._region = value
        self.info_display["Region"] = value

    @property
    def ssh_key(self):
        """
        Property. The current ssh key.
        """
        return self._ssh_key

    @ssh_key.setter
    def ssh_key(self, value):
        self._ssh_key = value
        self.info_display["SSH Key"] = value
        self.info_display["Default SSH username"] = (
            self.config["default_ssh_usernames"][value]
            if value in self.config["default_ssh_usernames"]
            else ""
        )

    def get_keypair_association(self, keypair_id):
        """
        Fetches the keypair association of a keypair from the configuration.

        Parameters
        ----------
        keypair_id : str
            The ID of the keypair being queried.

        Returns
        -------
        str
            The SSH key the keypair is associated with.
        """
        if keypair_id in self.config["keypair_associations"]:
            return self.config["keypair_associations"][keypair_id]
        return ""

    def set_keypair_association(self, keypair_id, key_name):
        """
        Sets the association between a keypair and an SSH key.

        Parameters
        ----------
        keypair_id : str
            The ID of the keypair being associated.
        key_name : str
            The name of the SSH key associated with the keypair.
        """
        self.config["keypair_associations"][keypair_id] = key_name
        self.config.write_config()

    def textedit(self, value):
        """
        Open the text editor command specified in the editor_command configuration option to edit a text. Returns when the called command
        exits.

        Parameters
        ----------
        value : str
            The initial text to edit in the text editor.

        Returns
        -------
        str
            The edited text.
        """
        editor = self.config["editor_command"]
        with tempfile.NamedTemporaryFile("w", delete=False) as temp:
            temp_file = temp.name
            temp.write(value)
            temp.close()
        try:
            self.ui.unraw(subprocess.run, ["bash", "-c", editor.format(temp_file)])
            with open(temp_file, "r", encoding="utf-8") as temp:
                return temp.read()
        finally:
            os.unlink(temp.name)

    def get_default_dashboard_layout(self):
        """
        Returns the default configured dashboard layout.

        Returns
        -------
        list(list(str))
            A list of rows of dashboard block class names.
        """
        if "default_dashboard_layout" in self.config:
            return self.config["default_dashboard_layout"]
        return [["Blank", "Blank"], ["Blank", "Blank"]]

    def get_context_dashboard_layout(self):
        """
        Returns the configured dashboard layout for the currently selected context.

        Returns
        -------
        list(list(str))
            A list of rows of dashboard block class names.
        """
        if self.context in self.config["dashboard_layouts"]:
            return self.config["dashboard_layouts"][self.context]
        return self.get_default_dashboard_layout()



"""
This module defines classes which allow viewing plain text or json data.
"""
import pyperclip
from pygments.lexers.data import JsonLexer
from pygments.token import Token

from .color import Color, ColorBlackOnGold, ColorGold, Palette8Bit
from .common import Commons
from .control import HotkeyControl


class NullHighlighter:
    """
    Syntax highlighter for text browsers which does not highlight syntax.
    """

    def __call__(self, browser, lines):
        return [[(line, browser.color)] for line in lines]


class JsonHighlighter:
    """
    Syntax highlighter for text browsers displaying json.
    """

    def _get_scheme_color(self, browser, color_name):
        try:
            color = browser.scheme["colors"][color_name]
        except AttributeError:
            return browser.color
        except KeyError:
            return browser.color
        return Color(Palette8Bit(), color["foreground"], background=color["background"])

    def __call__(self, browser, lines):
        joined = "\n".join(lines)
        lexer = JsonLexer(tabsize=2, stripnl=False, stripall=False)
        tokens = lexer.get_tokens(joined)
        current_line = []
        result_lines = []
        for token in tokens:
            token_color_name = (
                f"syntax_highlight_{str(token[0]).lower().replace('.', '_')}"
            )
            token_color = self._get_scheme_color(browser, token_color_name)
            if token[0] in Token.Text:
                brk = token[1].split("\n")
                while len(brk) > 1:
                    current_line.append((brk[0], token_color))
                    result_lines.append(current_line)
                    current_line = []
                    brk = brk[1:]
                current_line.append((brk[0], token_color))
            else:
                current_line.append((token[1], token_color))
        result_lines.append(current_line)
        return result_lines


class TextBrowser(HotkeyControl):
    """
    Text browsers allow the display of large blobs of text which can be optionally syntax highlighted.

    Attributes
    ----------
    scheme : awsc.config.scheme.Scheme
        Reference to the scheme being used by the browser. Since this module has no access to the configuration, this needs to be explicitly
        passed.
    color : awsc.termui.color.Color
        Default color used for text within the browser.
    filtered_color : awsc.termui.color.Color
        Color used for highlighting text when using search.
    lines : list
        A raw list of lines to display, without syntax highlighting applied.
    display_lines : list
        A list of lines to display with syntax highlighting applied. Acts as a cache that is cleared when new raw lines are added.
    top : int
        The amount of rows by which the control is scrolled down.
    left : int
        The amount of characters by which the control is scrolled right.
    wrap : bool
        If set, wraps lines that would go off the right edge. Disables scrolling horizontally.
    _filter : str
        Search string.
    filter_positions : dict
        A cache of all positions within the control where matches for the search are found.
    """

    def __init__(
        self,
        *args,
        color=ColorGold,
        filtered_color=ColorBlackOnGold,
        scheme=None,
        syntax_highlighting=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.entries = []

        self.add_hotkey("KEY_DOWN", self.scroll_down)
        self.add_hotkey("KEY_UP", self.scroll_up)
        self.add_hotkey("KEY_LEFT", self.scroll_left)
        self.add_hotkey("KEY_RIGHT", self.scroll_right)
        self.add_hotkey("KEY_END", self.end)
        self.add_hotkey("KEY_HOME", self.home)
        self.add_hotkey("KEY_PGUP", self.pgup)
        self.add_hotkey("KEY_PGDOWN", self.pgdown)
        self.add_hotkey("c", self.copy_contents, "Copy")
        self.add_hotkey("w", self.toggle_wrap, "Toggle Wrap")
        self.scheme = scheme
        self.color = color
        self.filtered_color = filtered_color
        self.lines = []
        self.display_lines = []
        self.top = 0
        self.left = 0
        self._filter_line = 0
        self.wrap = False
        self._filter = None
        self.filter_positions = {}
        self._current_filter_match = []
        self.syntax_highlighter = (
            JsonHighlighter() if syntax_highlighting else NullHighlighter()
        )

    def _raw(self, line_part):
        if isinstance(line_part, tuple):
            return line_part[0]
        return line_part

    def rawlines(self):
        """
        Generates a list of raw lines.

        Returns
        -------
        list(str)
            A list of lines in the text browser.
        """
        return [self.raw(i) for i in range(len(self.lines))]

    def raw(self, line=None):
        """
        Generates the raw output of the text browser.

        Parameters
        ----------
        line : int, optional
            If provided, returns the line with this index, otherwise, returns the text of the entire text browser.

        Returns
        -------
        str
            The raw text of the entire browser if no line index is provided, otherwise, the raw text of a single line.
        """
        if line is None:
            return "\n".join(self.rawlines())
        return "".join(
            [self._raw(self.lines[line][i]) for i in range(len(self.lines[line]))]
        )

    def add_text(self, text):
        """
        Adds raw text to the text browser.

        Parameters
        ----------
        text : str
            The text to append to the existing text in the browser.
        """
        add = text.split("\n")
        self.lines.extend(add)
        self.display_lines = self.syntax_highlighter(self, self.lines)
        Commons.UIInstance.dirty = True

    def clear(self):
        """
        Empties out the text browser.
        """
        Commons.UIInstance.dirty = True
        self.lines = []
        self.display_lines = []
        self.filter_positions = {}
        self.filter = None
        self.top = 0
        self.left = 0

    @property
    def filter(self):
        """
        Property. If changed, searches for that particular snippet of text within the browser. Set to None to stop searching.
        """
        return self._filter

    @filter.setter
    def filter(self, value):
        Commons.UIInstance.dirty = True
        refilter = False
        if self._filter != value:
            self._filter_line = -1
            self.filter_positions = {}
            self._current_filter_match = []
            refilter = True

        self._filter = value

        if value is None or value == "":
            return

        if refilter:
            for i in range(len(self.lines)):
                rawline = self.raw(i)
                processed_line = rawline.split(self._filter)
                pos = 0
                for j in range(1, len(processed_line)):
                    pos += len(processed_line[j - 1])
                    if i not in self.filter_positions:
                        self.filter_positions[i] = []
                    self.filter_positions[i].append(pos)
                if len(processed_line) > 1:
                    self._current_filter_match.append(i)

        for elem in self._current_filter_match:
            if elem > self._filter_line:
                self._filter_line = elem
                self.top = elem
                return
        if len(self._current_filter_match) > 0:
            self.top = self._current_filter_match[0]
            self._filter_line = self.top

    def toggle_wrap(self, *args):
        """
        Hotkey callback for toggling the wrap flag.
        """
        self.wrap = not self.wrap
        if self.wrap:
            self.left = 0

    def scroll_down(self, *args):
        """
        Hotkey callback for scrolling down within the browser by a single line.
        """
        if self.top < len(self.lines) - 1:
            self.top += 1

    def scroll_up(self, *args):
        """
        Hotkey callback for scrolling up within the browser by a single line.
        """
        if self.top > 0:
            self.top -= 1

    def scroll_right(self, *args):
        """
        Hotkey callback for scrolling right within the browser by a single character column.
        """
        if self.wrap:
            return
        self.left += 1

    def scroll_left(self, *args):
        """
        Hotkey callback for scrolling left within the browser by a single character column.
        """
        if self.wrap:
            return
        if self.left > 0:
            self.left -= 1

    def end(self, *args):
        """
        Hotkey callback for scrolling to bring the final line in view within the browser.
        """
        corners = self.corners
        y0 = corners[1][0] + (0 if self.border is None else 1)
        y1 = corners[1][1] - (0 if self.border is None else 1)
        height = y1 - y0 + 1
        limit = len(self.lines) - height if not self.wrap else len(self.lines) - 1
        self.top = max(limit, 0)

    def home(self, *args):
        """
        Hotkey callback for jumping to the top of the text browser.
        """
        self.top = 0

    def pgup(self, *args):
        """
        Hotkey callback for scrolling up within the browser by a single page.
        """
        corners = self.corners
        y0 = corners[1][0] + (0 if self.border is None else 1)
        y1 = corners[1][1] - (0 if self.border is None else 1)
        height = y1 - y0 + 1
        self.top = max(self.top - height, 0)

    def pgdown(self, *args):
        """
        Hotkey callback for scrolling down within the browser by a single page.
        """
        corners = self.corners
        y0 = corners[1][0] + (0 if self.border is None else 1)
        y1 = corners[1][1] - (0 if self.border is None else 1)
        height = y1 - y0 + 1
        limit = len(self.lines) - height if not self.wrap else len(self.lines) - 1
        self.top = max(min(limit, self.top + height), 0)

    def copy_contents(self, *args):
        """
        Hotkey callback for copying the contents of the text browser to the system clipboard.
        """
        pyperclip.copy(self.raw())

    def paint(self):
        super().paint()
        ((x0, x1), (y, y1)) = self.inner
        for i, line in enumerate(self.display_lines):
            if i < self.top:
                continue
            skip_chars = self.left
            x = x0
            end = False
            pos = skip_chars
            next_filter_pos = -1
            next_filter_last_pos = -1
            nfp_idx = 0
            if i in self.filter_positions:
                for filter_position in self.filter_positions[i]:
                    if filter_position >= pos:
                        next_filter_pos = filter_position
                        next_filter_last_pos = filter_position + len(self._filter) - 1
                        break
                    if filter_position + len(self._filter) - 1 >= pos:
                        next_filter_pos = filter_position
                        next_filter_last_pos = filter_position + len(self._filter) - 1
                    else:
                        nfp_idx += 1
            for elem in line:
                buf = self._raw(elem)
                if skip_chars > len(buf):
                    skip_chars -= len(buf)
                    continue
                if skip_chars > 0:
                    buf = buf[skip_chars:]
                    skip_chars = 0
                color = self.color if not isinstance(elem, tuple) else elem[1]

                while len(buf) > 0:
                    space = x1 - x + 1
                    text = buf[:space]
                    buf = buf[space:]
                    while len(text) > 0:
                        if next_filter_pos <= pos <= next_filter_last_pos >= 0:
                            filt = next_filter_last_pos - pos + 1
                            if filt >= len(text):
                                Commons.UIInstance.print(
                                    text, xy=(x, y), color=self.filtered_color
                                )
                                x += len(text)
                                pos += len(text)
                                text = ""
                            else:
                                Commons.UIInstance.print(
                                    text[:filt], xy=(x, y), color=self.filtered_color
                                )
                                text = text[filt:]
                                x += filt
                                pos += filt
                                nfp_idx += 1
                                if len(self.filter_positions[i]) > nfp_idx:
                                    next_filter_pos = self.filter_positions[i][nfp_idx]
                                    next_filter_last_pos = (
                                        next_filter_pos + len(self._filter) - 1
                                    )
                                else:
                                    next_filter_pos = -1
                                    next_filter_last_pos = -1
                        elif 0 <= next_filter_pos < pos + len(text):
                            unfilt = next_filter_pos - pos
                            if unfilt > 0:
                                Commons.UIInstance.print(
                                    text[:unfilt], xy=(x, y), color=color
                                )
                                text = text[unfilt:]
                                x += unfilt
                                pos += unfilt
                            filt = next_filter_last_pos - next_filter_pos + 1
                            if filt >= len(text):
                                Commons.UIInstance.print(
                                    text, xy=(x, y), color=self.filtered_color
                                )
                                x += len(text)
                                pos += len(text)
                                text = ""
                            else:
                                Commons.UIInstance.print(
                                    text[:filt], xy=(x, y), color=self.filtered_color
                                )
                                text = text[filt:]
                                x += filt
                                pos += filt
                                nfp_idx += 1
                                if len(self.filter_positions[i]) > nfp_idx:
                                    next_filter_pos = self.filter_positions[i][nfp_idx]
                                    next_filter_last_pos = (
                                        next_filter_pos + len(self._filter) - 1
                                    )
                                else:
                                    next_filter_pos = -1
                                    next_filter_last_pos = -1
                        else:
                            Commons.UIInstance.print(text, xy=(x, y), color=color)
                            pos += len(text)
                            x += len(text)
                            text = ""
                    if x > x1:
                        y += 1
                        x = x0
                        if not self.wrap:
                            end = True

                    if end or y > y1:
                        break
            if not end:
                y += 1
                x = x0
            if y > y1:
                break
